import doctest

from insights.parsers import virsh_list_all
from insights.tests import context_wrap


BLANK = """
""".strip()

NO_RESULT = """
 Id    Name                           State
----------------------------------------------------
""".strip()

OUTPUT = """
 Id    Name                           State
----------------------------------------------------
 2     rhel7.4                        running
 4     rhel7.0                        paused
 -     centos6.8-router               shut off
 -     cfme-5.7.13                    shut off
 -     cfme-rhos-5.9.0.15             shut off
 -     fedora-24-kernel               shut off
 -     fedora-saio_fedoraSaio         shut off
 -     fedora24-misc                  shut off
 -     freebsd11.0                    shut off
 -     guixSD                         shut off
 -     miq-gap-1                      shut off
 -     rhel7.2                        shut off
 -     RHOSP10                        shut off
""".strip()


def test_virsh_output():
    output = virsh_list_all.VirshListAll(context_wrap(OUTPUT))
    assert len(output.search(state='shut off')) == 11
    assert len(output.search(id=None)) == 11
    assert len(output.search(id=2)) == 1
    assert output.search(name='rhel7.4') == [{'state': 'running', 'id': 2, 'name': 'rhel7.4'}]
    assert output.get_vm_state('rhel7.0') == 'paused'
    assert output.get_vm_state('rhel9.0') is None
    assert ('cfme' in output) is False
    assert ('cfme-5.7.13' in output) is True


def test_virsh_output_no_vms():
    output = virsh_list_all.VirshListAll(context_wrap(NO_RESULT))
    assert output.fields == []
    assert output.cols == []
    assert output.keywords == []
    assert output.get_vm_state('NORHEL') is None


def test_virsh_output_blank():
    output = virsh_list_all.VirshListAll(context_wrap(BLANK))
    assert output.fields == []
    assert output.cols == []
    assert output.keywords == []
    assert output.get_vm_state('NORHEL') is None


def test_virsh_list_all_documentation():
    failed_count, tests = doctest.testmod(
        virsh_list_all,
        globs={'output': virsh_list_all.VirshListAll(context_wrap(OUTPUT))}
    )
    assert failed_count == 0



from ...parsers import hammer_task_list
from ...tests import context_wrap

import doctest

hammer_task_list_csv_data = '''
ID,Name,Owner,Started at,Ended at,State,Result,Task action,Task errors
92b732ea-7423-4644-8890-80e054f1799a,,foreman_api_admin,2016/11/11 07:18:32,2016/11/11 07:18:34,stopped,success,Refresh repository,""
e9cb6455-a433-467e-8404-7d01bd726689,,foreman_api_admin,2016/11/11 07:18:28,2016/11/11 07:18:31,stopped,success,Refresh repository,""
e30f3e7e-c023-4380-9594-337fdc4967e4,,foreman_api_admin,2016/11/11 07:18:24,2016/11/11 07:18:28,stopped,success,Refresh repository,""
3197f6a1-891f-4f42-9e4d-92c83c3ed035,,foreman_api_admin,2016/11/11 07:18:20,2016/11/11 07:18:24,stopped,success,Refresh repository,""
22169621-7175-411c-86be-46b4254a4e77,,foreman_api_admin,2016/11/11 07:18:16,2016/11/11 07:18:19,stopped,success,Refresh repository,""
f111e8f7-c956-470b-abb6-2e436ecd5866,,foreman_api_admin,2016/11/11 07:18:14,2016/11/11 07:18:16,stopped,success,Refresh repository,""
dfc702ea-ce46-427c-8a07-43e2a68e1320,,foreman_api_admin,2016/11/11 07:18:12,2016/11/11 07:18:14,stopped,success,Refresh repository,""
e8cac892-e666-4f2c-ab97-2be298da337e,,foreman_api_admin,2016/11/11 07:18:09,2016/11/11 07:18:12,stopped,success,Refresh repository,""
e6c1e1b2-a29d-4fd0-891e-e736dc9b7150,,,2016/11/11 07:14:06,2016/11/12 05:10:17,stopped,success,Listen on candlepin events,""
44a42c49-3038-4cae-8067-4d1cc305db05,,,2016/11/11 07:11:44,2016/11/11 07:12:47,stopped,success,Listen on candlepin events,""
72669288-54ac-41ba-a3b2-314a2c81f438,,,2016/11/11 06:57:15,2016/11/11 07:07:03,stopped,success,Listen on candlepin events,""
1314c91e-19d6-4d71-9bca-31db0df0aad2,,foreman_admin,2016/11/11 06:55:59,2016/11/11 06:55:59,stopped,error,Update for host sat62disc.example.org,"There was an issue with the backend service candlepin: 404 Resource Not Found, There was an issue with the backend service candlepin: 404 Resource Not Found"
303ef924-9845-4267-a705-194a4ebfbcfb,,foreman_admin,2016/11/11 06:55:58,2016/11/11 06:55:58,stopped,error,Package Profile Update,500 Internal Server Error
cffa5990-23ba-49f5-828b-ae0c77e8257a,,foreman_admin,2016/11/11 06:55:53,2016/11/11 06:55:56,stopped,error,Update for host sat62disc.example.org,"There was an issue with the backend service candlepin: 404 Resource Not Found, There was an issue with the backend service candlepin: 404 Resource Not Found"
07780e8f-dd81-49c4-a792-c4d4d162eb10,,foreman_admin,2016/11/11 06:55:50,2016/11/11 06:55:51,stopped,error,Update for host sat62disc.example.org,"There was an issue with the backend service candlepin: 404 Resource Not Found, There was an issue with the backend service candlepin: 404 Resource Not Found"
749a17a1-a8cb-46f0-98f6-017576481df8,,foreman_admin,2016/11/11 06:51:28,2016/11/11 06:51:29,stopped,error,Update for host sat62disc.example.org,"There was an issue with the backend service candlepin: 404 Resource Not Found, There was an issue with the backend service candlepin: 404 Resource Not Found"
d8f41819-b492-46e5-b0e3-ead3b4b6810c,,foreman_admin,2016/11/11 06:51:22,2016/11/11 06:51:28,stopped,error,Package Profile Update,500 Internal Server Error
'''


def test_HTL_doc_examples():
    tasks = hammer_task_list.HammerTaskList(context_wrap(hammer_task_list_csv_data))
    globs = {
        'tasks': tasks
    }
    failed, tested = doctest.testmod(hammer_task_list, globs=globs)
    assert failed == 0



"""
Mlx4Port - command ``find /sys/bus/pci/devices/*/mlx4_port[0-9] -print -exec cat {} \;``
========================================================================================

This module provides processing for the output of the
``find /sys/bus/pci/devices/*/mlx4_port[0-9] -print -exec cat {} \;`` command.

Sample output of this command looks like::

    /sys/bus/pci/devices/0000:0c:00.0/mlx4_port1
    ib
    /sys/bus/pci/devices/0000:0c:00.0/mlx4_port2
    eth

The ``port_val`` method is to return a dictionary contains mlx4_port name and
its content as the parsing result::

{'mlx4_port2': 'eth', 'mlx4_port1': 'ib'}

Examples:
    >>> mlx4_port_content = '''
    ... /sys/bus/pci/devices/0000:0c:00.0/mlx4_port1
    ... ib
    ... /sys/bus/pci/devices/0000:0c:00.0/mlx4_port2
    ... eth
    ... '''.strip()
    >>> from insights.tests import context_wrap
    >>> from insights.parsers.mlx4_port import Mlx4Port
    >>> shared = {Mlx4Port: Mlx4Port(context_wrap(mlx4_port_content))}
    >>> mlx4_port_results = Mlx4Port(context_wrap(mlx4_port_content))
    >>> mlx4_port_results.port_val
    {'mlx4_port2': 'eth', 'mlx4_port1': 'ib'}
"""

from collections import defaultdict
from .. import parser, CommandParser
from insights.specs import Specs


@parser(Specs.mlx4_port)
class Mlx4Port(CommandParser):
    """
    Parse the output of the command:

    ``find /sys/bus/pci/devices/*/mlx4_port[0-9] -print -exec cat {} \;``.
    """
    @property
    def port_val(self):
        """
        dict: Returns the dictionary of `mlx4 port` name and its content
        """
        return self._mapping

    def parse_content(self, content):
        self._mapping = defaultdict(list)
        port_name = None
        for line in content:
            if line.startswith('/sys/'):
                port_name = line.split('/')[6]
            elif port_name:
                self._mapping[port_name].append(line.strip())
        self._mapping = dict([(port, "\n".join(values)) for port, values in self._mapping.items()])



from insights.core.remote_resource import RemoteResource, CachedRemoteResource
from insights.tests.mock_web_server import TestMockServer
import sys
import pytest

GOOD_PAYLOAD = b'Successful return from Mock Service'
NOT_FOUND = b'{"error":{"code": "404", "message": "Not Found"}}'


class TestRemoteResource(TestMockServer):

    # Test RemoteResource
    def test_get_remote_resource(self):
        rr = RemoteResource()

        url = 'http://localhost:{port}/mock/'.format(port=self.server_port)
        rtn = rr.get(url)
        assert GOOD_PAYLOAD in rtn.content

    # Test RemoteResource not found
    def test_get_remote_resource_not_found(self):
        rr = RemoteResource()

        url = 'http://localhost:{port}/moc/'.format(port=self.server_port)
        rtn = rr.get(url)
        assert rtn.content == NOT_FOUND

    # Test CachedRemoteResource not cached
    def test_get_cached_remote_resource(self):
        crr = CachedRemoteResource()

        url = 'http://localhost:{port}/mock/'.format(port=self.server_port)
        rtn = crr.get(url)
        assert GOOD_PAYLOAD in rtn.content

    # Test CachedRemoteResource returns cached response
    @pytest.mark.skipif(sys.version_info < (2, 7), reason="CacheControl requires python 2.7 or higher")
    def test_get_cached_remote_resource_cached(self):
        crr = CachedRemoteResource()

        url = 'http://localhost:{port}/mock/'.format(port=self.server_port)
        rtn = crr.get(url)
        cont_1 = rtn.content
        rtn = crr.get(url)
        cont_2 = rtn.content
        assert cont_1 == cont_2

    # Test CachedRemoteResource not found
    def test_get_cached_remote_resource_not_found(self):
        crr = CachedRemoteResource()

        url = 'http://localhost:{port}/moc/'.format(port=self.server_port)
        rtn = crr.get(url)
        assert rtn.content == NOT_FOUND



from __future__ import print_function
import functools
import json
import logging
import os
import shutil
import sys
import six

from insights.client import InsightsClient
from insights.client.config import InsightsConfig
from insights.client.constants import InsightsConstants as constants
from insights.client.support import InsightsSupport
from insights.client.utilities import validate_remove_file
from insights.client.schedule import get_scheduler

logger = logging.getLogger(__name__)


def phase(func):
    @functools.wraps(func)
    def _f():
        try:
            config = InsightsConfig().load_all()
        except ValueError as e:
            sys.stderr.write('ERROR: ' + str(e) + '\n')
            sys.exit(constants.sig_kill_bad)
        client = InsightsClient(config)
        if config.debug:
            logger.info("Core path: %s", os.path.dirname(__file__))
        try:
            func(client, config)
        except Exception:
            logger.exception("Fatal error")
            sys.exit(1)
        else:
            sys.exit(0)  # Exit gracefully
    return _f


def get_phases():
    return [{
        'name': 'pre_update',
        'run_as_root': True
    }, {
        'name': 'update',
        'run_as_root': True
    }, {
        'name': 'post_update',
        'run_as_root': True
    }, {
        'name': 'collect_and_output',
        'run_as_root': True
    }]


@phase
def pre_update(client, config):
    if config.version:
        logger.info(constants.version)
        sys.exit(constants.sig_kill_ok)

    # validate the remove file
    if config.validate:
        if validate_remove_file(config.remove_file):
            sys.exit(constants.sig_kill_ok)
        else:
            sys.exit(constants.sig_kill_bad)

    # handle cron stuff
    if config.enable_schedule:
        # enable automatic scheduling
        logger.debug('Updating config...')
        updated = get_scheduler(config).set_daily()
        if updated:
            logger.info('Automatic scheduling for Insights has been enabled.')
        sys.exit(constants.sig_kill_ok)

    if config.disable_schedule:
        # disable automatic schedling
        updated = get_scheduler(config).remove_scheduling()
        if updated:
            logger.info('Automatic scheduling for Insights has been disabled.')
        if not config.register:
            sys.exit(constants.sig_kill_ok)

    # delete someday
    if config.analyze_container:
        logger.debug('Not scanning host.')
        logger.debug('Scanning image ID, tar file, or mountpoint.')

    # test the insights connection
    if config.test_connection:
        logger.info("Running Connection Tests...")
        rc = client.test_connection()
        if rc == 0:
            sys.exit(constants.sig_kill_ok)
        else:
            sys.exit(constants.sig_kill_bad)

    if config.support:
        support = InsightsSupport(config)
        support.collect_support_info()
        sys.exit(constants.sig_kill_ok)

    if config.diagnosis:
        remediation_id = None
        if config.diagnosis is not True:
            remediation_id = config.diagnosis
        resp = client.get_diagnosis(remediation_id)
        if not resp:
            sys.exit(constants.sig_kill_bad)
        print(json.dumps(resp))
        sys.exit(constants.sig_kill_ok)


@phase
def update(client, config):
    client.update()
    if config.payload:
        logger.debug('Uploading a payload. Bypassing rules update.')
        return
    client.update_rules()


@phase
def post_update(client, config):
    # create a machine id first thing. we'll need it for all uploads
    logger.debug('Machine ID: %s', client.get_machine_id())
    logger.debug("CONFIG: %s", config)

    # -------delete everything below this line-------
    if config.legacy_upload:
        if config.status:
            reg_check = client.get_registration_status()
            for msg in reg_check['messages']:
                logger.info(msg)
            if reg_check['status']:
                sys.exit(constants.sig_kill_ok)
            else:
                sys.exit(constants.sig_kill_bad)

        # put this first to avoid conflicts with register
        if config.unregister:
            if client.unregister():
                sys.exit(constants.sig_kill_ok)
            else:
                sys.exit(constants.sig_kill_bad)

        if config.offline:
            logger.debug('Running client in offline mode. Bypassing registration.')
            return

        if config.analyze_container:
            logger.debug(
                'Running client in container mode. Bypassing registration.')
            return

        if config.display_name and not config.register:
            # setting display name independent of registration
            if client.set_display_name(config.display_name):
                if 'display_name' in config._cli_opts:
                    # only exit on success if it was invoked from command line
                    sys.exit(constants.sig_kill_ok)
            else:
                sys.exit(constants.sig_kill_bad)

        reg = client.register()
        if reg is None:
            # API unreachable
            logger.info('Running connection test...')
            client.test_connection()
            sys.exit(constants.sig_kill_bad)
        elif reg is False:
            # unregistered
            sys.exit(constants.sig_kill_bad)
        if config.register:
            if (not config.disable_schedule and
               get_scheduler(config).set_daily()):
                logger.info('Automatic scheduling for Insights has been enabled.')
        return
    # -------delete everything above this line-------

    if config.offline:
        logger.debug('Running client in offline mode. Bypassing registration.')
        return

    # --payload short circuits registration check
    if config.payload:
        logger.debug('Uploading a specified archive. Bypassing registration.')
        return

    # check registration status before anything else
    reg_check = client.get_registration_status()
    if reg_check is None:
        sys.exit(constants.sig_kill_bad)

    # --status
    if config.status:
        if reg_check:
            logger.info('This host is registered.')
            sys.exit(constants.sig_kill_ok)
        else:
            logger.info('This host is unregistered.')
            sys.exit(constants.sig_kill_bad)

    # put this first to avoid conflicts with register
    if config.unregister:
        if reg_check:
            logger.info('Unregistration is not currently available.')
            sys.exit(constants.sig_kill_bad)
        else:
            logger.info('This host is not registered, unregistration is not applicable.')
            sys.exit(constants.sig_kill_bad)

    # halt here if unregistered
    if not reg_check and not config.register:
        logger.info('This host has not been registered. '
                    'Use --register to register this host.')
        sys.exit(constants.sig_kill_bad)

    # --force-reregister, clear machine-id
    if config.reregister:
        reg_check = False
        client.clear_local_registration()

    # --register was called
    if config.register:
        # don't actually need to make a call to register() since
        #   system creation and upload are a single event on the platform
        if reg_check:
            logger.info('This host has already been registered.')
        if (not config.disable_schedule and
           get_scheduler(config).set_daily()):
            logger.info('Automatic scheduling for Insights has been enabled.')

    # set --display-name independent of register
    # only do this if set from the CLI. normally display_name is sent on upload
    if 'display_name' in config._cli_opts and not config.register:
        if client.set_display_name(config.display_name):
            sys.exit(constants.sig_kill_ok)
        else:
            sys.exit(constants.sig_kill_bad)


@phase
def collect_and_output(client, config):
    if config.payload:
        insights_archive = config.payload
    else:
        try:
            insights_archive = client.collect()
        except RuntimeError as e:
            logger.error(e)
            sys.exit(constants.sig_kill_bad)
        config.content_type = 'application/vnd.redhat.advisor.collection+tgz'

    if not insights_archive:
        sys.exit(constants.sig_kill_bad)
    if config.to_stdout:
        with open(insights_archive, 'rb') as tar_content:
            if six.PY3:
                sys.stdout.buffer.write(tar_content.read())
            else:
                shutil.copyfileobj(tar_content, sys.stdout)
    else:
        resp = None
        if not config.no_upload:
            try:
                resp = client.upload(payload=insights_archive, content_type=config.content_type)
            except IOError as e:
                logger.error(str(e))
                sys.exit(constants.sig_kill_bad)
            except ValueError as e:
                logger.error(str(e))
                sys.exit(constants.sig_kill_bad)
        else:
            logger.info('Archive saved at %s', insights_archive)
        if resp:
            if config.to_json:
                print(json.dumps(resp))

            if not config.payload:
                # delete the archive
                if config.keep_archive:
                    logger.info('Insights archive retained in ' + insights_archive)
                else:
                    client.delete_archive(insights_archive, delete_parent_dir=True)
    client.delete_cached_branch_info()

    # rotate eggs once client completes all work successfully
    try:
        client.rotate_eggs()
    except IOError:
        message = ("Failed to rotate %s to %s" %
                   (constants.insights_core_newest,
                    constants.insights_core_last_stable))
        logger.debug(message)
        raise IOError(message)



from insights.parsers import yum
from insights.parsers.yum import YumRepoList
from insights.parsers import SkipException
from insights.tests import context_wrap
import doctest
import pytest

YUM_REPOLIST_CONTENT = """
Loaded plugins: product-id, search-disabled-repos, subscription-manager
repo id                                         repo name                status
rhel-7-server-rpms/7Server/x86_64               Red Hat Enterprise Linux 10415
rhel-7-server-satellite-6.1-rpms/x86_64         Red Hat Satellite 6.1 (f   660
rhel-7-server-satellite-capsule-6.1-rpms/x86_64 Red Hat Satellite Capsul   265
rhel-server-rhscl-7-rpms/7Server/x86_64         Red Hat Software Collect  4571
repolist: 15911

""".strip()

YUM_REPOLIST_CONTENT_EUS = """
Loaded plugins: product-id, rhnplugin, security, subscription-manager
Updating certificate-based repositories.
repo id                              repo name                            status
clone-6u5-server-x86_64              clone-6u5-server-x86_64              3,787
rhel-x86_64-server-6.2.aus           Red Hot Enterprise Linu              5,509
repolist: 3,787

""".strip()

YUM_REPOLIST_CONTENT_MISSING_STATUS = """
Loaded plugins: product-id, rhnplugin, security, subscription-manager
Updating certificate-based repositories.
repo id                              repo name                            status
clone-6u5-server-x86_64              clone-6u5-server-x86_64
"""

YUM_REPOLIST_CONTENT_OUT_OF_DATE = """
Loaded plugins: product-id, search-disabled-repos, subscription-manager
Repodata is over 2 weeks old. Install yum-cron? Or run: yum makecache fast
repo id                              repo name                                          status
!rhel-7-server-extras-rpms/x86_64    Red Hat Enterprise Linux 7 Server - Extras (RPMs)  877
!rhel-7-server-rpms/7Server/x86_64   Red Hat Enterprise Linux 7 Server (RPMs)           20,704
repolist: 21,581
"""

YUM_REPOLIST_DOC = """
Loaded plugins: langpacks, product-id, search-disabled-repos, subscription-manager
repo id                                             repo name                                                                                                    status
rhel-7-server-e4s-rpms/x86_64                       Red Hat Enterprise Linux 7 Server - Update Services for SAP Solutions (RPMs)                                 12,250
rhel-ha-for-rhel-7-server-e4s-rpms/x86_64           Red Hat Enterprise Linux High Availability (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)         272
rhel-ha-for-rhel-7-server-rpms/x86_64               Red Hat Enterprise Linux High Availability (for RHEL 7 Server) (RPMs)                                           225
rhel-sap-hana-for-rhel-7-server-e4s-rpms/x86_64     RHEL for SAP HANA (for RHEL 7 Server) Update Services for SAP Solutions (RPMs)                                   21
repolist: 12,768
"""


def test_yum_repolist():
    repo_list = YumRepoList(context_wrap(YUM_REPOLIST_CONTENT))
    assert len(repo_list) == 4
    assert repo_list[0] == {"id": "rhel-7-server-rpms/7Server/x86_64",
                            "name": "Red Hat Enterprise Linux",
                            "status": "10415"}
    assert 'rhel-7-server-rpms/7Server/x86_64' in repo_list
    assert repo_list['rhel-7-server-rpms/7Server/x86_64'] == repo_list[0]
    assert repo_list.eus == []


def test_eus():
    repo_list = YumRepoList(context_wrap(YUM_REPOLIST_CONTENT_EUS))
    assert len(repo_list) == 2
    assert repo_list[0] == {"id": "clone-6u5-server-x86_64",
                            "name": "clone-6u5-server-x86_64",
                            "status": "3,787"}
    assert repo_list.eus == ["6.2.aus"]


def test_rhel_repos():
    repo_list = YumRepoList(context_wrap(YUM_REPOLIST_CONTENT))
    assert len(repo_list.rhel_repos) == 4
    assert set(repo_list.rhel_repos) == set(['rhel-7-server-rpms',
                                             'rhel-7-server-satellite-6.1-rpms',
                                             'rhel-7-server-satellite-capsule-6.1-rpms',
                                             'rhel-server-rhscl-7-rpms'])


def test_rhel_repos_missing_status():
    with pytest.raises(SkipException) as se:
        YumRepoList(context_wrap(YUM_REPOLIST_CONTENT_MISSING_STATUS))
    assert 'Incorrect line:' in str(se)


def test_rhel_repos_empty():
    with pytest.raises(SkipException) as se:
        YumRepoList(context_wrap(''))
    assert 'No repolist.' in str(se)


def test_rhel_repos_out_of_date():
    repo_list = YumRepoList(context_wrap(YUM_REPOLIST_CONTENT_OUT_OF_DATE))
    assert len(repo_list) == 2
    assert set(repo_list.rhel_repos) == set(['rhel-7-server-extras-rpms',
                                             'rhel-7-server-rpms'])


def test_invalid_get_type():
    repo_list = YumRepoList(context_wrap(YUM_REPOLIST_CONTENT))
    assert repo_list[YumRepoList] is None


def test_doc_examples():
    env = {
            'repolist': YumRepoList(context_wrap(YUM_REPOLIST_DOC)),
          }
    failed, total = doctest.testmod(yum, globs=env)
    assert failed == 0



import pytest

from insights.tests import context_wrap
from insights.parsers.alternatives import AlternativesOutput, JavaAlternatives
from insights.core import ParseException

ALT_MTA = """
mta - status is auto.
 link currently points to /usr/sbin/sendmail.postfix
/usr/sbin/sendmail.postfix - priority 30
 slave mta-mailq: /usr/bin/mailq.postfix
 slave mta-newaliases: /usr/bin/newaliases.postfix
Current `best' version is /usr/sbin/sendmail.postfix.
"""

DUPLICATED_STATUS_LINE = """
mta - status is auto.
Nonsense line that should be ignored
mta - status is auto.
"""

MISSING_STATUS_LINE = """
 link currently points to /usr/sbin/sendmail.postfix
/usr/sbin/sendmail.postfix - priority 30
 slave mta-mailq: /usr/bin/mailq.postfix
 slave mta-newaliases: /usr/bin/newaliases.postfix
Current `best' version is /usr/sbin/sendmail.postfix.
"""


def test_mta_alternatives():
    mtas = AlternativesOutput(context_wrap(ALT_MTA))

    assert hasattr(mtas, 'program')
    assert mtas.program == 'mta'
    assert hasattr(mtas, 'status')
    assert mtas.status == 'auto'
    assert hasattr(mtas, 'link')
    assert mtas.link == '/usr/sbin/sendmail.postfix'
    assert hasattr(mtas, 'best')
    assert mtas.best == '/usr/sbin/sendmail.postfix'

    assert hasattr(mtas, 'paths')
    assert isinstance(mtas.paths, list)
    assert len(mtas.paths) == 1

    for i in ('path', 'priority', 'slave'):
        assert i in mtas.paths[0]


def test_failure_modes():
    # Duplicate status line raises ParseException
    with pytest.raises(ParseException, match='Program line for mta'):
        alts = AlternativesOutput(context_wrap(DUPLICATED_STATUS_LINE))
        assert alts.program is None

    # Missing status line results in no data
    alts = AlternativesOutput(context_wrap(MISSING_STATUS_LINE))
    for i in (alts.program, alts.status, alts.link, alts.best):
        assert i is None
    assert alts.paths == []


alter_java = """
java - status is auto.
 link currently points to /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java
/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java - priority 16091
 slave ControlPanel: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel
 slave keytool: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/keytool
 slave policytool: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/policytool
 slave rmid: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/rmid
 slave rmiregistry: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/rmiregistry
 slave tnameserv: /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/tnameserv
 slave jre_exports: /usr/lib/jvm-exports/jre-1.6.0-ibm.x86_64
 slave jre: /usr/lib/jvm/jre-1.6.0-ibm.x86_64
/usr/lib/jvm/jre-1.4.2-gcj/bin/java - priority 1420
 slave ControlPanel: (null)
 slave keytool: /usr/lib/jvm/jre-1.4.2-gcj/bin/keytool
 slave policytool: (null)
 slave rmid: (null)
 slave rmiregistry: /usr/lib/jvm/jre-1.4.2-gcj/bin/rmiregistry
 slave tnameserv: (null)
 slave jre_exports: /usr/lib/jvm-exports/jre-1.4.2-gcj
 slave jre: /usr/lib/jvm/jre-1.4.2-gcj
Current `best' version is /usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java.
""".strip()

alter_no_java = """
/usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/java - priority 170079
 slave ControlPanel: (null)
 slave keytool: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/keytool
 slave orbd: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/orbd
 slave pack200: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/pack200
 slave policytool: (null)
 slave rmid: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/rmid
 slave rmiregistry: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/rmiregistry
 slave servertool: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/servertool
 slave tnameserv: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/tnameserv
 slave unpack200: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/unpack200
 slave jre_exports: /usr/lib/jvm-exports/jre-1.7.0-openjdk.x86_64
 slave jre: /usr/lib/jvm/jre-1.7.0-openjdk.x86_64
 slave java.1.gz: /usr/share/man/man1/java-java-1.7.0-openjdk.1.gz
 slave keytool.1.gz: /usr/share/man/man1/keytool-java-1.7.0-openjdk.1.gz
 slave orbd.1.gz: /usr/share/man/man1/orbd-java-1.7.0-openjdk.1.gz
 slave pack200.1.gz: /usr/share/man/man1/pack200-java-1.7.0-openjdk.1.gz
 slave rmid.1.gz: /usr/share/man/man1/rmid-java-1.7.0-openjdk.1.gz
 slave rmiregistry.1.gz: /usr/share/man/man1/rmiregistry-java-1.7.0-openjdk.1.gz
 slave servertool.1.gz: /usr/share/man/man1/servertool-java-1.7.0-openjdk.1.gz
 slave tnameserv.1.gz: /usr/share/man/man1/tnameserv-java-1.7.0-openjdk.1.gz
 slave unpack200.1.gz: /usr/share/man/man1/unpack200-java-1.7.0-openjdk.1.gz
Current `best' version is /usr/lib/jvm/jre-1.7.0-openjdk.x86_64/bin/java.
""".strip()


def test_class_no_java():
    java = JavaAlternatives(context_wrap(alter_no_java))
    for i in (java.program, java.status, java.link, java.best):
        assert i is None
    assert java.paths == []


def test_class_has_java():
    java = JavaAlternatives(context_wrap(alter_java))

    assert java.program == 'java'
    assert java.status == 'auto'
    assert java.link == '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
    assert java.best == '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'

    assert isinstance(java.paths, list)
    assert len(java.paths) == 2

    for i in ('path', 'priority', 'slave'):
        assert i in java.paths[0]

    assert java.paths[0]['path'] == '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/java'
    assert java.paths[0]['priority'] == 16091
    assert isinstance(java.paths[0]['slave'], dict)
    assert sorted(java.paths[0]['slave'].keys()) == sorted([
        'ControlPanel', 'keytool', 'policytool', 'rmid', 'rmiregistry',
        'tnameserv', 'jre_exports', 'jre'
    ])
    assert java.paths[0]['slave']['ControlPanel'] == '/usr/lib/jvm/jre-1.6.0-ibm.x86_64/bin/ControlPanel'

    for i in ('path', 'priority', 'slave'):
        assert i in java.paths[1]

    assert java.paths[1]['path'] == '/usr/lib/jvm/jre-1.4.2-gcj/bin/java'
    assert java.paths[1]['priority'] == 1420
    assert isinstance(java.paths[1]['slave'], dict)
    assert sorted(java.paths[1]['slave'].keys()) == sorted([
        'ControlPanel', 'keytool', 'policytool', 'rmid', 'rmiregistry',
        'tnameserv', 'jre_exports', 'jre'
    ])
    assert java.paths[1]['slave']['ControlPanel'] == '(null)'



import doctest

from insights.parsers import systemctl_show
from insights.parsers.systemctl_show import SystemctlShowCinderVolume
from insights.parsers.systemctl_show import SystemctlShowHttpd
from insights.parsers.systemctl_show import SystemctlShowMariaDB
from insights.parsers.systemctl_show import SystemctlShowPulpCelerybeat
from insights.parsers.systemctl_show import SystemctlShowPulpResourceManager
from insights.parsers.systemctl_show import SystemctlShowPulpWorkers
from insights.parsers.systemctl_show import SystemctlShowQpidd
from insights.parsers.systemctl_show import SystemctlShowQdrouterd
from insights.parsers.systemctl_show import SystemctlShowSmartpdc
from insights.tests import context_wrap


SYSTEMCTL_SHOW_EXAMPLES = """
WatchdogUSec=0
WatchdogTimestamp=Thu 2018-01-11 14:22:33 CST
WatchdogTimestampMonotonic=105028136
StartLimitInterval=10000000
StartLimitBurst=5
StartLimitAction=none
FailureAction=none
PermissionsStartOnly=no
RootDirectoryStartOnly=no
RemainAfterExit=no
GuessMainPID=yes
MainPID=2810
ControlPID=0
FileDescriptorStoreMax=0
StatusErrno=0
Result=success
ExecMainStartTimestamp=Thu 2018-01-11 14:22:33 CST
ExecMainStartTimestampMonotonic=105028117
ExecMainExitTimestampMonotonic=0
ExecMainPID=2811
LimitNOFILE=4096
"""


def test_systemctl_show_cinder_volume():
    context = SystemctlShowCinderVolume(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_mariadb():
    context = SystemctlShowMariaDB(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_pulp_workers():
    context = SystemctlShowPulpWorkers(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_pulp_resource_manager():
    context = SystemctlShowPulpResourceManager(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_pulp_celerybeat():
    context = SystemctlShowPulpCelerybeat(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_httpd():
    context = SystemctlShowHttpd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_qpidd():
    context = SystemctlShowQpidd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_qdrouterd():
    context = SystemctlShowQdrouterd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_smartpdc():
    context = SystemctlShowSmartpdc(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    assert context["LimitNOFILE"] == "4096"
    assert len(context.data) == 21


def test_systemctl_show_doc_examples():
    env = {
        'systemctl_show_cinder_volume': SystemctlShowCinderVolume(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_mariadb': SystemctlShowMariaDB(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_pulp_workers': SystemctlShowPulpWorkers(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_pulp_resource_manager': SystemctlShowPulpResourceManager(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_pulp_celerybeat': SystemctlShowPulpCelerybeat(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_httpd': SystemctlShowHttpd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_qpidd': SystemctlShowQpidd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_qdrouterd': SystemctlShowQdrouterd(context_wrap(SYSTEMCTL_SHOW_EXAMPLES)),
        'systemctl_show_smartpdc': SystemctlShowSmartpdc(context_wrap(SYSTEMCTL_SHOW_EXAMPLES))
    }
    failed, total = doctest.testmod(systemctl_show, globs=env)
    assert failed == 0



from insights.parsers import cpuset_cpus
from insights.tests import context_wrap


CPUSET_CPU = """
0,2-4,7
""".strip()


def test_init_process_cgroup():
    cpusetinfo = cpuset_cpus.CpusetCpus(context_wrap(CPUSET_CPU))
    assert cpusetinfo.cpu_set == ["0", "2", "3", "4", "7"]
    assert cpusetinfo.cpu_number == 5



"""
Partitions - file ``/proc/partitions``
======================================

This parser reads the ``/proc/partitions`` file, which contains
partition block allocation information.

"""
from insights import Parser, parser
from insights.parsers import SkipException, parse_delimited_table
from insights.specs import Specs


@parser(Specs.partitions)
class Partitions(Parser, dict):
    """A class for parsing the ``/proc/partitions`` file.

    Sample input::

        major minor  #blocks  name

           3     0   19531250 hda
           3     1     104391 hda1
           3     2   19422585 hda2
         253     0   22708224 dm-0
         253     1     524288 dm-1

    Examples:
       >>> type(partitions_info)
       <class 'insights.parsers.partitions.Partitions'>
       >>> 'hda' in partitions_info
       True
       >>> partitions_info['dm-0'].get('major')
       '253'
       >>> sorted(partitions_info['hda'].items(), key=lambda x: x[0])
       [('blocks', '19531250'), ('major', '3'), ('minor', '0'), ('name', 'hda')]

    Attributes:
        partitions (:obj:`dict`): Dictionary with each partition name as index and
            its information from the block allocation table.

    Raises:
        SkipException: When input is empty.

    """
    def __init__(self, *args, **kwargs):
        super(Partitions, self).__init__(*args, **kwargs)
        self.update(self.partitions)

    def parse_content(self, content):
        if not content:
            raise SkipException('Empty content')

        self.partitions = {}

        self.partitions = dict(
            (row['name'], row)
            for row in parse_delimited_table(
                    content,
                    heading_ignore=['major', 'minor'],
                    header_substitute=[('#blocks', 'blocks')]
            )
            if 'name' in row
        )



from insights.parsers.redhat_release import RedhatRelease
from insights.parsers.ps import PsAuxcww
from insights.parsers.httpd_V import HttpdV as HV
from insights.parsers.httpd_V import HttpdWorkerV as HWV
from insights.parsers.httpd_V import HttpdEventV as HEV
from insights.combiners.httpd_V import HttpdV
from insights.tests import context_wrap, RHEL6, RHEL7
from insights import SkipComponent
import pytest

HTTPDV1 = """
Server version: Apache/2.2.15 (Unix)
Server built:   Feb  4 2016 02:44:09
Server's Module Magic Number: 20051115:25
Server loaded:  APR 1.3.9, APR-Util 1.3.9
Compiled using: APR 1.3.9, APR-Util 1.3.9
Architecture:   64-bit
Server MPM:     Prefork
  threaded:     no
    forked:     yes (variable process count)
Server compiled with....
 -D APACHE_MPM_DIR="server/mpm/prefork"
 -D APR_HAS_SENDFILE
 -D APR_HAS_MMAP
 -D APR_HAVE_IPV6 (IPv4-mapped addresses enabled)
 -D APR_USE_SYSVSEM_SERIALIZE
 -D APR_USE_PTHREAD_SERIALIZE
 -D SINGLE_LISTEN_UNSERIALIZED_ACCEPT
 -D APR_HAS_OTHER_CHILD
 -D AP_HAVE_RELIABLE_PIPED_LOGS
 -D DYNAMIC_MODULE_LIMIT=128
 -D HTTPD_ROOT="/etc/httpd"
 -D SUEXEC_BIN="/usr/sbin/suexec"
 -D DEFAULT_PIDLOG="run/httpd.pid"
 -D DEFAULT_SCOREBOARD="logs/apache_runtime_status"
 -D DEFAULT_LOCKFILE="logs/accept.lock"
 -D DEFAULT_ERRORLOG="logs/error_log"
 -D AP_TYPES_CONFIG_FILE="conf/mime.types"
 -D SERVER_CONFIG_FILE="conf/httpd.conf"
""".strip()

HTTPDV2 = """
Server version: Apache/2.4.6 (Red Hat Enterprise Linux)
Server built:   Aug  3 2016 08:33:27
Server's Module Magic Number: 20120211:24
Server loaded:  APR 1.4.8, APR-UTIL 1.5.2
Compiled using: APR 1.4.8, APR-UTIL 1.5.2
Architecture:   64-bit
Server MPM:     worker
  threaded:     no
    forked:     yes (variable process count)
Server compiled with....
 -D APR_HAS_MMAP
 -D APR_HAVE_IPV6 (IPv4-mapped addresses enabled)
 -D APR_USE_SYSVSEM_SERIALIZE
 -D APR_USE_PTHREAD_SERIALIZE
 -D SINGLE_LISTEN_UNSERIALIZED_ACCEPT
 -D APR_HAS_OTHER_CHILD
 -D AP_HAVE_RELIABLE_PIPED_LOGS
 -D DYNAMIC_MODULE_LIMIT=256
 -D HTTPD_ROOT="/etc/httpd"
 -D SUEXEC_BIN="/usr/sbin/suexec"
 -D DEFAULT_PIDLOG="/run/httpd/httpd.pid"
 -D DEFAULT_SCOREBOARD="logs/apache_runtime_status"
 -D DEFAULT_ERRORLOG="logs/error_log"
 -D AP_TYPES_CONFIG_FILE="conf/mime.types"
 -D SERVER_CONFIG_FILE="conf/httpd.conf"
""".strip()

HTTPDV3 = """
Server version: Apache/2.4.6 (Red Hat Enterprise Linux)
Server built:   Aug  3 2016 08:33:27
Server's Module Magic Number: 20120211:24
Server loaded:  APR 1.4.8, APR-UTIL 1.5.2
Compiled using: APR 1.4.8, APR-UTIL 1.5.2
Architecture:   64-bit
Server MPM:     event
  threaded:     no
    forked:     yes (variable process count)
Server compiled with....
 -D APR_HAS_MMAP
""".strip()

PS_WORKER = """
USER  PID %CPU %MEM    VSZ   RSS TTY STAT START   TIME COMMAND
root   41  0.0  0.0  21452  1536 ?   Ss   Mar09   0:01 httpd.worker
root   75  0.0  0.0      0     0 ?   S    Mar09   0:00 [kthreadd]
""".strip()

PS_EVENT = """
USER  PID %CPU %MEM    VSZ   RSS TTY STAT START   TIME COMMAND
root   41  0.0  0.0  21452  1536 ?   Ss   Mar09   0:01 httpd.event
root   75  0.0  0.0      0     0 ?   S    Mar09   0:00 [kthreadd]
""".strip()


def test_httpd_V_RHEL6():
    hv1 = HV(context_wrap(HTTPDV1))
    hv2 = HWV(context_wrap(HTTPDV2, path='httpd.worker_-V'))
    hv3 = HEV(context_wrap(HTTPDV3, path='httpd.event_-V'))
    ps = PsAuxcww(context_wrap(PS_WORKER))
    rh = RedhatRelease(context_wrap(RHEL6))
    result = HttpdV(rh, ps, hv1, hv3, hv2)
    assert result["Server MPM"] == "worker"
    assert result["Server version"] == "apache/2.4.6 (red hat enterprise linux)"
    assert result["forked"] == "yes (variable process count)"
    assert "APR_HAVE_IPV6" in result['Server compiled with']
    assert result['Server compiled with']['APR_HAS_MMAP'] is True
    assert result['Server compiled with']['APR_HAVE_IPV6'] == "IPv4-mapped addresses enabled"
    assert result['Server compiled with']['DEFAULT_PIDLOG'] == "/run/httpd/httpd.pid"

    ps = PsAuxcww(context_wrap(PS_EVENT))
    result = HttpdV(rh, ps, hv1, hv3, hv2)
    assert result["Server MPM"] == "event"
    assert result["Server version"] == "apache/2.4.6 (red hat enterprise linux)"
    assert result["forked"] == "yes (variable process count)"
    assert "APR_HAVE_IPV6" not in result['Server compiled with']
    assert result['Server compiled with']['APR_HAS_MMAP'] is True


def test_httpd_V_RHEL7():
    hv1 = HV(context_wrap(HTTPDV1))
    hv2 = HWV(context_wrap(HTTPDV2, path='httpd.worker_-V'))
    hv3 = HEV(context_wrap(HTTPDV3, path='httpd.event_-V'))
    ps = PsAuxcww(context_wrap(PS_WORKER))
    rh = RedhatRelease(context_wrap(RHEL7))
    result = HttpdV(rh, ps, hv1, hv3, hv2)
    assert result["Server MPM"] == "prefork"
    assert result["Server version"] == "apache/2.2.15 (unix)"
    assert result["forked"] == "yes (variable process count)"
    assert "APR_HAVE_IPV6" in result['Server compiled with']
    assert result['Server compiled with']['APR_HAS_MMAP'] is True
    assert result['Server compiled with']['APR_HAVE_IPV6'] == "IPv4-mapped addresses enabled"
    assert result['Server compiled with']['DEFAULT_PIDLOG'] == "run/httpd.pid"


def test_httpd_V_failed():
    hv1 = HV(context_wrap(HTTPDV1))
    hv2 = HWV(context_wrap(HTTPDV2, path='httpd.worker_-V'))
    ps = PsAuxcww(context_wrap(PS_EVENT))
    rh = RedhatRelease(context_wrap(RHEL6))
    with pytest.raises(SkipComponent) as sc:
        HttpdV(rh, ps, hv1, None, hv2)
    assert "Unable to get the valid `httpd -V` command" in str(sc)



import sys
from insights.core.dr import load_components


def get_archives(module_name, system_filter):
    load_components(module_name)
    m = sys.modules[module_name]
    for sub_m in m.__all__:
        demo_submodule = sys.modules[".".join([module_name, sub_m])]
        if hasattr(demo_submodule, "demo"):
            for a in demo_submodule.demo:
                if system_filter in a.name:
                    yield a



from __future__ import print_function
import hashlib
import os
import sys
from .tool import TestArchive, Transform, MultiArchive
from insights.tests import InputData
from insights.core import load_package

ARCHIVES = [
    TestArchive("base_rhel67", base_archive="rhel6.7"),
    TestArchive("insights_heartbeat", transforms=[
        Transform("hostname").replace("insights-heartbeat-9cd6f607-6b28-44ef-8481-62b0e7773614")
    ], machine_id=hashlib.sha224("TEST-HEARTBEAT-RHAI2").hexdigest())
]


DEFAULT_DEST = "./integration_tests"
console = False
archive_map = {}
for archive in ARCHIVES:
    archive_map[archive.name] = archive


def build_all(archives, dest="./archives", clean=False):
    if not os.path.exists(dest):
        os.mkdir(dest)
    files = []
    for archive in archives:
        files.append(archive.build(dest, force=clean))
        print("%s: %s" % (archive.name, archive.machine_id))
    return files


def build(archive_name, dest="."):
    return archive_map[archive_name].build(dest)


def get_integration_archive_name(module_name, input_data, i):
    invalid_chars = [" ", "/", "$", "%", "*", "(", ")", ":", "%"]
    if isinstance(input_data, InputData) and input_data.name:
        name = input_data.name.lower()
        for invalid_char in invalid_chars:
            name = name.replace(invalid_char, "_")
    elif isinstance(input_data, list):
        name = "multinode-%d" % i
    else:
        name = str(i)
    return "_".join([module_name, name])


def build_integration_test_archive(module_name, i, test_tuple, dest, machine_id=None):
    module, test_func, input_data, expected_response = test_tuple
    test_name = get_integration_archive_name(module.__name__, input_data, i)
    try:
        if isinstance(input_data, list):
            archive = MultiArchive(test_name, machine_id=machine_id)
        else:
            archive = TestArchive(test_name, machine_id=machine_id)
        archive.apply_input_data(input_data)
        archive_path = archive.build(dest)
    except Exception:
        print("X", end="")
    else:
        if console:
            print("m" if isinstance(input_data, list) else ".", end="")
            sys.stdout.flush()
        return {
            "module": module_name,
            "test_func": test_func,
            "input_data": input_data,
            "expected_response": expected_response,
            "archive_path": archive_path,
            "test_archive": archive
        }


def build_integration_test_archives(package, dest=DEFAULT_DEST, machine_id=None):
    if not os.path.exists(dest):
        os.mkdir(dest)
    for func in get_integration_generators(package):
        module_name = func.__module__
        if console:
            print("\nBuilding {0} ".format(module_name), end="")
        for i, test_tuple in enumerate(func()):
            archive = build_integration_test_archive(module_name, i, test_tuple, dest, machine_id)
            if archive:
                yield archive
    if console:
        print()


def get_integration_generators(package):
    load_package(package)
    from insights.tests import ARCHIVE_GENERATORS
    for func in sorted(ARCHIVE_GENERATORS, key=lambda f: f.__module__):
        if not func.slow:
            yield func



"""
passenger-status command
========================
This module provides processing for the ``passenger-status`` command using the
following parsers:
"""
from insights.parsers import ParseException
from insights.specs import Specs

from .. import CommandParser, get_active_lines, parser


@parser(Specs.passenger_status)
class PassengerStatus(CommandParser):
    """
    Parse the passenger-status command output.

    Produces a simple dictionary of keys and values from the command output
    contents, stored in the ``data`` attribute.

    Sample command output::

        Version : 4.0.18
        Date    : 2018-10-23 15:42:04 +0800
        Instance: 1265
        ----------- General information -----------
        Max pool size : 12
        Processes     : 2
        Requests in top-level queue : 0

        ----------- Application groups -----------
        /usr/share/foreman#default:
        App root: /usr/share/foreman
        Requests in queue: 192
        * PID: 30131   Sessions: 1       Processed: 991     Uptime: 2h 9m 8s
        CPU: 3%      Memory  : 562M    Last used: 1h 53m 51s
        * PID: 32450   Sessions: 1       Processed: 966     Uptime: 2h 8m 15s
        CPU: 4%      Memory  : 463M    Last used: 1h 48m 17
        * PID: 4693    Sessions: 1       Processed: 939     Uptime: 2h 6m 32s
        CPU: 3%      Memory  : 470M    Last used: 1h 50m 48

        /etc/puppet/rack#default:
        App root: /etc/puppet/rack
        Requests in queue: 0
        * PID: 21934   Sessions: 1       Processed: 380     Uptime: 1h 33m 34s
        CPU: 1%      Memory  : 528M    Last used: 1h 29m 4
        * PID: 26194   Sessions: 1       Processed: 544     Uptime: 1h 31m 34s
        CPU: 2%      Memory  : 490M    Last used: 1h 23m 5
        * PID: 32384   Sessions: 1       Processed: 36      Uptime: 1h 0m 29s
        CPU: 0%      Memory  : 561M    Last used: 1h 0m 3s


    Example data structure produced::

        {
            'Version': '4.0.18',
            'Date': '2018-10-23 15',
            'Instance': '1265',
            'Max pool size': '12',
            'Processes': '2',
            'Requests in top-level queue': '0',
            'foreman_default': {
                'App root': '/usr/share/foreman',
                'Requests in queue': '192',
                'p_list': [
                    {
                        'Uptime': '2h 9m 8s', 'Processed': '991', 'Sessions': '1', 'Memory': '562M',
                        'PID': '30131', 'CPU': '3%', 'Last used': '1h 53m 51s'
                    }, {
                        'Uptime': '2h 8m 15s', 'Processed': '966', 'Sessions': '1', 'Memory': '463M',
                        'PID': '32450', 'CPU': '4%', 'Last used': '1h 48m 17'
                    }, {
                        'Uptime': '2h 6m 32s', 'Processed': '939', 'Sessions': '1', 'Memory': '470M',
                        'PID': '4693', 'CPU': '3%', 'Last used': '1h 50m 48'
                    }
                ]
            },
            'rack_default': {
                'App root': '/etc/puppet/rack',
                'Requests in queue': '0',
                'p_list': [
                    {
                        'Uptime': '1h 33m 34s', 'Processed': '380', 'Sessions': '1', 'Memory': '528M',
                        'PID': '21934', 'CPU': '1%', 'Last used': '1h 29m 4'
                    }, {
                        'Uptime': '1h 31m 34s', 'Processed': '544', 'Sessions': '1', 'Memory': '490M',
                        'PID': '26194', 'CPU': '2%', 'Last used': '1h 23m 5'
                    }, {
                        'Uptime': '1h 0m 29s', 'Processed': '36', 'Sessions': '1', 'Memory': '561M',
                        'PID': '32384', 'CPU': '0%', 'Last used': '1h 0m 3s'
                    }
                ]
            }
        }


    Examples:
        >>> passenger_status.data["Version"]
        '4.0.18'
        >>> 'rack_default' in passenger_status.data
        True
        >>> len(passenger_status.data['foreman_default']['p_list'])
        3

    Raises:
        ParseException: When there is no useful data or the input content is
            empty, or does contain the header line.
    """

    def parse_content(self, content):
        content = get_active_lines(content, '-----------')
        if len(content) <= 1:
            raise ParseException("Input content is empty or there is no useful parsed data.")

        if not content or not content[0].startswith('Version'):
            raise ParseException("Cannot find the header line.")
        returndic = {}
        group_bit = None
        for line in content:
            if not group_bit and ":" in line and not line.endswith(':'):
                returndic[line.split(":")[0].strip()] = line.split(":")[1].strip()
                continue

            if line.startswith(("/usr/share/foreman#default", "/etc/puppet/rack#default")):
                # set the group which the following pid will belong to
                group_bit = line.strip(':').split('/')[-1].replace('#', '_')
                returndic[group_bit] = {}
                continue

            if group_bit:
                if line.strip().startswith(("* PID:", "CPU")):
                    if line.strip().startswith("* PID:"):
                        p_dic = {}
                        # if this is the first process of this group the create the p_lists key
                        if 'p_list' not in returndic[group_bit]:
                            returndic[group_bit]['p_list'] = []
                        returndic[group_bit]['p_list'].append(p_dic)
                    linesplit = line.split("    ")
                    for i in linesplit:
                        if ':' in i:
                            si1, si2 = i.split(':', 1)
                            p_dic[si1.strip(' \t*')] = si2.strip()
                elif ':' in line:
                    # set the entry to the current group
                    returndic[group_bit][line.split(":")[0].strip()] = line.split(":")[1].strip()

        self.data = returndic



import doctest
import pytest
from insights.parsers import SkipException
from ...parsers import subscription_manager_list
from ...tests import context_wrap


subscription_manager_list_consumed_in_docs = '''
+-------------------------------------------+
   Consumed Subscriptions
+-------------------------------------------+
Subscription Name: Red Hat Enterprise Linux Server, Premium (1-2 sockets) (Up to 1 guest)
Provides:          Oracle Java (for RHEL Server)
                   Red Hat Software Collections Beta (for RHEL Server)
                   Red Hat Enterprise Linux Server
                   Red Hat Beta
SKU:               RH0155783S
Contract:          12345678
Account:           1000001
Serial:            0102030405060708090
Pool ID:           8a85f981477e5284014783abaf5d4dcd
Active:            True
Quantity Used:     1
Service Level:     PREMIUM
Service Type:      L1-L3
Status Details:    Subscription is current
Subscription Type: Standard
Starts:            11/14/14
Ends:              07/06/15
System Type:       Physical
'''

subscription_manager_list_installed_in_docs = '''
+-------------------------------------------+
Installed Product Status
+-------------------------------------------+
Product Name:   Red Hat Software Collections (for RHEL Server)
Product ID:     201
Version:        2
Arch:           x86_64
Status:         Subscribed
Status Details:
Starts:         04/27/15
Ends:           04/27/16

Product Name:   Red Hat Enterprise Linux Server
Product ID:     69
Version:        7.1
Arch:           x86_64
Status:         Subscribed
Status Details:
Starts:         04/27/15
Ends:           04/27/16
'''

subscription_manager_repos_list_enabled_test_data = '''
+----------------------------------------------------------+
    Available Repositories in /etc/yum.repos.d/redhat.repo
+----------------------------------------------------------+
Repo ID:   rhel-7-server-ansible-2-rpms
Repo Name: Red Hat Ansible Engine 2 RPMs for Red Hat Enterprise Linux 7 Server
Repo URL:  https://cdn.redhat.com/content/dist/rhel/server/7/7Server/$basearch/ansible/2/os
Enabled:   1
'''

FACTS_LIST = """
cpu.core(s)_per_socket: 2
cpu.cpu(s): 2
cpu.cpu_socket(s): 1
cpu.thread(s)_per_core: 1
cpu.topology_source: kernel /sys cpu sibling lists
distribution.id: Maipo
distribution.name: Red Hat Enterprise Linux Server
distribution.version: 7.2
distribution.version.modifier: ga
dmi.baseboard.manufacturer: Oracle Corporation
dmi.baseboard.product_name: VirtualBox
dmi.baseboard.serial_number: 0
dmi.baseboard.version: 1.2
dmi.bios.address: 0xe0000
dmi.bios.relase_date: 12/01/2006
dmi.bios.rom_size: 128 KB
dmi.bios.runtime_size: 128 KB
dmi.bios.vendor: innotek GmbH
dmi.bios.version: VirtualBox
dmi.chassis.asset_tag: Not Specified
dmi.chassis.boot-up_state: Safe
dmi.chassis.lock: Not Present
dmi.chassis.manufacturer: Oracle Corporation
dmi.chassis.power_supply_state: Safe
dmi.chassis.security_status: None
dmi.chassis.serial_number: Not Specified
dmi.chassis.thermal_state: Safe
dmi.chassis.type: Other
dmi.chassis.version: Not Specified
dmi.system.family: Virtual Machine
dmi.system.manufacturer: innotek GmbH
dmi.system.product_name: VirtualBox
dmi.system.serial_number: 0
dmi.system.sku_number: Not Specified
dmi.system.uuid: e81897b5-94df-4979-a8e2-b496756b5cbc
dmi.system.version: 1.2
dmi.system.wake-up_type: Power Switch
lscpu.architecture: x86_64
lscpu.bogomips: 5587.10
lscpu.byte_order: Little Endian
lscpu.core(s)_per_socket: 2
lscpu.cpu(s): 2
lscpu.cpu_family: 6
lscpu.cpu_mhz: 2793.550
lscpu.cpu_op-mode(s): 32-bit, 64-bit
lscpu.hypervisor_vendor: KVM
lscpu.l1d_cache: 32K
lscpu.l1i_cache: 32K
lscpu.l2_cache: 256K
lscpu.l3_cache: 6144K
lscpu.model: 60
lscpu.model_name: Intel(R) Core(TM) i7-4810MQ CPU @ 2.80GHz
lscpu.numa_node(s): 1
lscpu.numa_node0_cpu(s): 0,1
lscpu.on-line_cpu(s)_list: 0,1
lscpu.socket(s): 1
lscpu.stepping: 3
lscpu.thread(s)_per_core: 1
lscpu.vendor_id: GenuineIntel
lscpu.virtualization_type: full
memory.memtotal: 1884064
memory.swaptotal: 520188
net.interface.enp0s3.ipv4_address: 10.0.2.15
net.interface.enp0s3.ipv4_broadcast: 10.0.2.255
net.interface.enp0s3.ipv4_netmask: 24
net.interface.enp0s3.ipv6_address.link: e80f::00:27ff:fe4a:c5efa
net.interface.enp0s3.ipv6_netmask.link: 64
net.interface.enp0s3.mac_address: AA:BB:CC:DD:EE:FF
net.interface.enp0s8.mac_address: AA:BB:CC:DD:EE:FF
net.interface.enp0s9.mac_address: AA:BB:CC:DD:EE:FF
net.interface.lo.ipv4_address: 127.0.0.1
net.interface.lo.ipv4_broadcast: Unknown
net.interface.lo.ipv4_netmask: 8
net.interface.lo.ipv6_address.host: ::1
net.interface.lo.ipv6_netmask.host: 128
network.hostname: rhel7-box
network.ipv4_address: 127.0.0.1
network.ipv6_address: ::1
system.certificate_version: 3.2
uname.machine: x86_64
uname.nodename: rhel7-box
uname.release: 3.10.0-327.el7.x86_64
uname.sysname: Linux
uname.version: #1 SMP Thu Oct 29 17:29:29 EDT 2015
virt.host_type: virtualbox, kvm
virt.is_guest: True
virt.uuid: 81897b5e-4df9-9794-8e2a-b496756b5cbc
""".strip()


subscription_manager_list_test_data = '''
+-------------------------------------------+
   Consumed Subscriptions
+-------------------------------------------+
Subscription Name: Red Hat Enterprise Linux Server, Premium (1-2 sockets) (Up to 1 guest)
Subscription Type: Standard
Starts:            17/2
'''

subscription_manager_list_no_installed_products = '''
No installed products to list
'''


def test_subscription_manager_list_exceptions():
    sml = subscription_manager_list.SubscriptionManagerListConsumed(
        context_wrap(subscription_manager_list_test_data)
    )
    assert len(sml.records) == 1
    rec0 = sml.records[0]
    assert 'Subscription Name' in rec0
    assert 'Subscription Type' in rec0
    assert 'Starts' in rec0
    assert rec0['Starts'] == '17/2'
    assert 'Starts timestamp' not in rec0

    sml = subscription_manager_list.SubscriptionManagerListInstalled(
        context_wrap(subscription_manager_list_no_installed_products)
    )
    assert sml.records == []


def test_all_facts():
    output = subscription_manager_list.SubscriptionManagerFactsList(context_wrap(FACTS_LIST))
    assert output['virt.uuid'] == '81897b5e-4df9-9794-8e2a-b496756b5cbc'
    assert output['net.interface.lo.ipv6_address.host'] == "::1"
    assert output['net.interface.enp0s9.mac_address'] == "AA:BB:CC:DD:EE:FF"
    assert output['net.interface.enp0s3.ipv6_address.link'] == "e80f::00:27ff:fe4a:c5efa"
    assert output['virt.host_type'] == "virtualbox, kvm"


def test_no_facts():
    with pytest.raises(SkipException) as ex:
        subscription_manager_list.SubscriptionManagerFactsList(context_wrap(''))
    assert "Empty content." in str(ex)


def test_subscription_manager_list_docs():
    env = {
        'repolist': subscription_manager_list.SubscriptionManagerReposListEnabled(context_wrap(subscription_manager_repos_list_enabled_test_data)),
        'installed': subscription_manager_list.SubscriptionManagerListInstalled(context_wrap(subscription_manager_list_installed_in_docs)),
        'consumed': subscription_manager_list.SubscriptionManagerListConsumed(context_wrap(subscription_manager_list_consumed_in_docs)),
        'facts': subscription_manager_list.SubscriptionManagerFactsList(context_wrap(FACTS_LIST))
    }
    failed, total = doctest.testmod(subscription_manager_list, globs=env)
    assert failed == 0



"""
IfCFG - files ``/etc/sysconfig/network-scripts/ifcfg-*``
========================================================

IfCFG is a parser for the network interface definition files in
``/etc/sysconfig/network-scripts``.  These are pulled into the network
scripts using ``source``, so they are mainly ``bash`` environment
declarations of the form **KEY=value**.  These are stored in the ``data``
property as a dictionary.  Quotes surrounding the value

Three options are handled differently:

* ``BONDING_OPTS`` is usually a quoted list of key=value arguments separated
  by spaces.
* ``TEAM_CONFIG`` and ``TEAM_PORT_CONFIG`` are treated as JSON stored as a
  single string.  Double quotes within the string are escaped using double
  back slashes, and these are removed so that the quoting is preserved.

Because this parser reads multiple files, the interfaces are stored as a
list within the parser and need to be iterated through in order to find
specific interfaces.

Sample configuration from a teamed interface in file ``/etc/sysconfig/network-scripts/ifcfg-team1``::

    DEVICE=team1
    DEVICETYPE=Team
    ONBOOT=yes
    NETMASK=255.255.252.0
    IPADDR=192.168.0.1
    TEAM_CONFIG='{"runner": {"name": "lacp", "active": "true", "tx_hash": ["eth", "ipv4"]}, "tx_balancer": {"name": "basic"}, "link_watch": {"name": "ethtool"}}'

Examples:

    >>> for nic in shared[IfCFG]: # Parser contains list of all interfaces
    ...     print 'NIC:', nic.iname
    ...     print 'IP address:', nic['IPADDR']
    ...     if 'TEAM_CONFIG' in nic:
    ...         print 'Team runner name:', nic['TEAM_CONFIG']['runner']['name']
    ...
    NIC: team1
    IP addresss: 192.168.0.1
    Team runner name: lacp

"""

import json
import re
from collections import OrderedDict
from .. import parser, get_active_lines, LegacyItemAccess, CommandParser
from insights.specs import Specs

JSON_FIELDS = ["TEAM_CONFIG", "TEAM_PORT_CONFIG"]

QUOTES = "\"'"

bond_mode_map = {
    'balance-rr': 0,
    'active-backup': 1,
    'balance-xor': 2,
    'broadcast': 3,
    '802.3ad': 4,
    'balance-tlb': 5,
    'balance-alb': 6
}


@parser(Specs.ifcfg)
class IfCFG(LegacyItemAccess, CommandParser):
    """
    Parse `ifcfg-` file,return a dict contain ifcfg config file info.
    "iface" key is interface name parse from file name
    `TEAM_CONFIG`, `TEAM_PORT_CONFIG` will return a dict with user config dict
    `BONDING_OPTS` also will return a dict

    Properties:
        ifname (str): The interface name as defined in the name of the file
            (i.e. the part after ``ifcfg-``).
    """

    def __init__(self, context):
        super(IfCFG, self).__init__(context)
        self.data["iface"] = context.path.rsplit("-", 1)[1]
        self.ifname = self.data['iface']
        self._has_empty_line = any(l.strip() == '' for l in context.content)

    def parse_content(self, content):
        self.data = {}
        for line in get_active_lines(content):
            if "=" not in line:
                continue
            key, value = line.split("=", 1)
            # Since keys are variable names in bash, stripping quotes and
            # spaces off them makes no sense.
            key = key.strip().strip(QUOTES).upper()

            # In some cases we want to know what the actual value-side
            # of the key is before dequoting and stripping.
            if key in ["DEVICE", "MASTER", "TEAM_MASTER", "BONDING_OPTS"]:
                self.data["raw_{0}_value".format(key.split('_')[0].lower())] = value
            if key != "DEVICE":
                value = value.strip().strip(QUOTES)
            if key in JSON_FIELDS:
                value = json.loads(value.replace("\\", ""))
            if key == "BONDING_OPTS":
                value_map = OrderedDict()
                value = re.sub(r'\s*=\s*', '=', value)
                for key_value_pair in value.split():
                    sub_key, sub_value = [
                        s.strip() for s in key_value_pair.split("=", 1)
                    ]
                    value_map[sub_key] = sub_value
                value = value_map
            self.data[key] = value

    @property
    def bonding_mode(self):
        """
        (int) the numeric value of bonding mode, or `None` if no bonding
        mode is found.
        """
        if "BONDING_OPTS" not in self or 'mode' not in self['BONDING_OPTS']:
            return None

        m = self["BONDING_OPTS"]["mode"]
        if m.isdigit():
            return int(m)
        if m in bond_mode_map:
            return bond_mode_map[m]
        return None

    @property
    def has_empty_line(self):
        """
        (bool) `True` if the file has empty line else `False`.
        """
        return self._has_empty_line



import pytest
from insights.parsers.hostname import Hostname
from insights.parsers.facter import Facter
from insights.parsers.systemid import SystemID
from insights.combiners.hostname import hostname
from insights.tests import context_wrap

HOSTNAME = "rhel7.example.com"
HOSTNAME_SHORT = "rhel7"
FACTS_FQDN = """
COMMAND> facter

architecture => x86_64
augeasversion => 1.1.0
facterversion => 1.7.6
filesystems => xfs
fqdn => ewa-satellite.example.com
hostname => ewa-satellite
""".strip()
FACTS_NO_FQDN = """
COMMAND> facter

architecture => x86_64
facterversion => 1.7.6
filesystems => xfs
""".strip()
SYSTEMID_PROFILE_NAME = '''
<?xml version="1.0"?>
<params>
<param>
<value><struct>
<member>
<name>username</name>
<value><string>testuser</string></value>
</member>
<member>
<name>profile_name</name>
<value><string>example_profile</string></value>
</member>
<member>
<name>system_id</name>
<value><string>ID-example</string></value>
</member>
<member>
<name>architecture</name>
<value><string>x86_64</string></value>
</member>
</struct></value>
</param>
</params>
'''.strip()
SYSTEMID_NO_PROFILE_NAME = '''
<?xml version="1.0"?>
<params>
<param>
<member>
<name>username</name>
<value><string>testuser</string></value>
</member>
<member>
<name>architecture</name>
<value><string>x86_64</string></value>
</member>
</param>
</params>
'''.strip()


def test_get_hostname():
    hn = Hostname(context_wrap(HOSTNAME))
    expected = (HOSTNAME, HOSTNAME_SHORT, 'example.com')
    result = hostname(hn, None, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]

    hn = Hostname(context_wrap(HOSTNAME_SHORT))
    expected = (HOSTNAME_SHORT, HOSTNAME_SHORT, '')
    result = hostname(hn, None, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_facter_hostname():
    hn = Facter(context_wrap(FACTS_FQDN))
    expected = ('ewa-satellite.example.com', 'ewa-satellite', 'example.com')
    result = hostname(None, hn, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_systemid_hostname():
    hn = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    expected = ('example_profile', 'example_profile', '')
    result = hostname(None, None, hn)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_all_hostname():
    hn = Hostname(context_wrap(HOSTNAME))
    fhn = Facter(context_wrap(FACTS_FQDN))
    shn = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    expected = (HOSTNAME, HOSTNAME_SHORT, 'example.com')
    result = hostname(hn, fhn, shn)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_hostname_raise():
    with pytest.raises(Exception):
        hn = Hostname(context_wrap(""))
        hostname(hn, None, None)



"""
Module responsible for scheduling Insights data collection in cron
"""
from __future__ import absolute_import
import os
import logging

from .constants import InsightsConstants as constants
from .utilities import run_command_get_output

APP_NAME = constants.app_name
logger = logging.getLogger(__name__)


def cron_source(config):
    extension = '-container' if config.analyze_container else ''
    return '/etc/%s/%s%s.cron' % (APP_NAME, APP_NAME, extension)


class InsightsSchedulerCron(object):

    def __init__(self, config, source=None, target='/etc/cron.daily/' + APP_NAME):
        self.target = target
        self.source = source if source else cron_source(config)

    @property
    def active(self):
        if os.path.exists(self.source):
            return os.path.isfile(self.target)
        return False

    def set_daily(self):
        logger.debug('Scheduling cron.daily')
        try:
            if not os.path.exists(self.target):
                os.symlink(self.source, self.target)
            return True
        except OSError:
            logger.exception('Could not link cron.daily')
            return False

    def remove_scheduling(self):
        logger.debug('Removing all cron tasks')
        try:
            # Remove symlinks in the case of rhel 6 running cron jobs
            if os.path.exists(self.target):
                os.remove(self.target)
            return True
        except OSError:
            logger.exception('Could not remove cron.daily')
            return False


class InsightsSchedulerSystemd(object):

    @property
    def active(self):
        try:
            systemctl_status = run_command_get_output('systemctl is-enabled insights-client.timer')
            return systemctl_status['status'] == 0
        except OSError:
            logger.exception('Could not get systemd status')
            return False

    def set_daily(self):
        logger.debug('Starting systemd timer')
        try:
            # Start timers in the case of rhel 7 running systemd
            systemctl_timer = run_command_get_output('systemctl start insights-client.timer')
            systemctl_timer = run_command_get_output('systemctl enable insights-client.timer')
            logger.debug("Starting Insights Client systemd timer.")
            logger.debug("Status: %s", systemctl_timer['status'])
            logger.debug("Output: %s", systemctl_timer['output'])
            return self.active
        except OSError:
            logger.exception('Could not start systemd timer')
            return False

    def remove_scheduling(self):
        logger.debug('Stopping all systemd timers')
        try:
            # Stop timers in the case of rhel 7 running systemd
            systemctl_timer = run_command_get_output('systemctl disable insights-client.timer')
            systemctl_timer = run_command_get_output('systemctl stop insights-client.timer')
            logger.debug("Stopping Insights Client systemd timer.")
            logger.debug("Status: %s", systemctl_timer['status'])
            logger.debug("Output: %s", systemctl_timer['output'])
            return not self.active
        except OSError:
            logger.exception('Could not stop systemd timer')
            return False


def get_scheduler(config, source=None, target='/etc/cron.daily/' + APP_NAME):
    source = source if source else cron_source(config)
    if os.path.exists(source):
        return InsightsSchedulerCron(config, source, target)
    else:
        return InsightsSchedulerSystemd()



"""
This module implements an inversion of control framework. It allows
dependencies among functions and classes to be declared with decorators and the
resulting dependency graphs to be executed.

A decorator used to declare dependencies is called a :class:`ComponentType`, a
decorated function or class is called a component, and a collection of
interdependent components is called a graph.

In the example below, ``needs`` is a :class:`ComponentType`, ``one``, ``two``,
and ``add`` are components, and the relationship formed by their dependencies
is a graph.

    .. code-block:: python

       from insights import dr

       class needs(dr.ComponentType):
           pass

       @needs()
       def one():
           return 1

       @needs()
       def two():
           return 2

       @needs(one, two)
       def add(a, b):
           return a + b

       results = dr.run(add)

Once all components have been imported, the graphs they form can be run. To
execute a graph, ``dr`` sorts its components into an order that guarantees
dependencies are tried before dependents. Components that raise exceptions are
considered invalid, and their dependents will not be executed. If a component
is skipped because of a missing dependency, its dependents also will not be
executed.

During evaluation, results are accumulated into an object called a
:class:`Broker`, which is just a fancy dictionary. Brokers can be inspected
after a run for results, exceptions, tracebacks, and execution times. You also
can register callbacks with a broker that get invoked after the attempted
execution of every component, so you can inspect it during an evaluation
instead of at the end.
"""
from __future__ import print_function

import inspect
import logging
import json
import os
import pkgutil
import re
import six
import sys
import time
import traceback

from collections import defaultdict
from functools import reduce as _reduce

from insights.contrib import importlib
from insights.contrib.toposort import toposort_flatten
from insights.util import defaults, enum, KeyPassingDefaultDict

log = logging.getLogger(__name__)

GROUPS = enum("single", "cluster")

MODULE_NAMES = {}
BASE_MODULE_NAMES = {}

TYPE_OBSERVERS = defaultdict(set)

COMPONENTS_BY_TYPE = defaultdict(set)
DEPENDENCIES = defaultdict(set)
DEPENDENTS = defaultdict(set)
COMPONENTS = defaultdict(lambda: defaultdict(set))

DELEGATES = {}
HIDDEN = set()
IGNORE = defaultdict(set)
ENABLED = defaultdict(lambda: True)


def set_enabled(component, enabled=True):
    """
    Enable a component for evaluation. If set to False, the component is
    skipped, and all components that require it will not execute. If component
    is a fully qualified name string of a callable object instead of the
    callable object itself, the component's module is loaded as a side effect
    of calling this function.

    Args:
        component (str or callable): fully qualified name of the component or
            the component object itself.
        enabled (bool): whether the component is enabled for evaluation.

    Returns:
        None
    """
    ENABLED[get_component(component) or component] = enabled


def is_enabled(component):
    """
    Check to see if a component is enabled.

    Args:
        component (callable): The component to check. The component must
            already be loaded.

    Returns:
        True if the component is enabled. False otherwise.
    """
    return ENABLED[component]


def get_delegate(component):
    return DELEGATES.get(component)


def add_ignore(c, i):
    IGNORE[c].add(i)


def hashable(v):
    try:
        hash(v)
    except:
        return False
    return True


def _get_from_module(name):
    mod, _, n = name.rpartition(".")
    if mod not in sys.modules:
        importlib.import_module(mod)
    return getattr(sys.modules[mod], n)


def _get_from_class(name):
    mod, _, n = name.rpartition(".")
    cls = _get_from_module(mod)
    return getattr(cls, n)


def _import_component(name):
    """
    Returns a class, function, or class method specified by the fully qualified
    name.
    """
    for f in (_get_from_module, _get_from_class):
        try:
            return f(name)
        except:
            pass
    log.debug("Couldn't load %s" % name)


COMPONENT_IMPORT_CACHE = KeyPassingDefaultDict(_import_component)


def get_component(name):
    """ Returns the class or function specified, importing it if necessary. """
    return COMPONENT_IMPORT_CACHE[name]


def _find_component(name):
    for d in DELEGATES:
        if get_name(d) == name:
            return d


COMPONENTS_BY_NAME = KeyPassingDefaultDict(_find_component)


def get_component_by_name(name):
    """
    Look up a component by its fully qualified name. Return None if the
    component hasn't been loaded.
    """
    return COMPONENTS_BY_NAME[name]


@defaults(None)
def get_component_type(component):
    return get_delegate(component).type


def get_components_of_type(_type):
    return COMPONENTS_BY_TYPE.get(_type)


@defaults(None)
def get_group(component):
    return get_delegate(component).group


def add_dependent(component, dep):
    DEPENDENTS[component].add(dep)


def get_dependents(component):
    return DEPENDENTS.get(component, set())


@defaults(set())
def get_dependencies(component):
    return get_delegate(component).get_dependencies()


def add_dependency(component, dep):
    get_delegate(component).add_dependency(dep)


class MissingRequirements(Exception):
    """
    Raised during evaluation if a component's dependencies aren't met.
    """
    def __init__(self, requirements):
        self.requirements = requirements
        super(MissingRequirements, self).__init__(requirements)


class SkipComponent(Exception):
    """
    This class should be raised by components that want to be taken out of
    dependency resolution.
    """
    pass


def get_name(component):
    """
    Attempt to get the string name of component, including module and class if
    applicable.
    """
    if six.callable(component):
        name = getattr(component, "__qualname__", component.__name__)
        return '.'.join([component.__module__, name])
    return str(component)


def get_simple_name(component):
    if six.callable(component):
        return component.__name__
    return str(component)


def get_metadata(component):
    """
    Return any metadata dictionary associated with the component. Defaults to
    an empty dictionary.
    """
    return get_delegate(component).metadata if component in DELEGATES else {}


def get_tags(component):
    """
    Return the set of tags associated with the component. Defaults to
    ``set()``.
    """
    return get_delegate(component).tags if component in DELEGATES else set()


def get_module_name(obj):
    try:
        return inspect.getmodule(obj).__name__
    except:
        return None


def get_base_module_name(obj):
    try:
        return get_module_name(obj).split(".")[-1]
    except:
        return None


def mark_hidden(component):
    global HIDDEN
    if isinstance(component, (list, set)):
        HIDDEN |= set(component)
    else:
        HIDDEN.add(component)


def is_hidden(component):
    return component in HIDDEN


def walk_tree(root, method=get_dependencies):
    for d in method(root):
        yield d
        for c in walk_tree(d, method=method):
            yield c


def walk_dependencies(root, visitor):
    """
    Call visitor on root and all dependencies reachable from it in breadth
    first order.

    Args:
        root (component): component function or class
        visitor (function): signature is `func(component, parent)`.  The
            call on root is `visitor(root, None)`.
    """
    def visit(parent, visitor):
        for d in get_dependencies(parent):
            visitor(d, parent)
            visit(d, visitor)

    visitor(root, None)
    visit(root, visitor)


def get_dependency_graph(component):
    """
    Generate a component's graph of dependencies, which can be passed to
    :func:`run` or :func:`run_incremental`.
    """
    if component not in DEPENDENCIES:
        raise Exception("%s is not a registered component." % get_name(component))

    if not DEPENDENCIES[component]:
        return {component: set()}

    graph = defaultdict(set)

    def visitor(c, parent):
        if parent is not None:
            graph[parent].add(c)

    walk_dependencies(component, visitor)

    graph = dict(graph)

    # Find all items that don't depend on anything.
    extra_items_in_deps = _reduce(set.union, graph.values(), set()) - set(graph.keys())

    # Add empty dependencies where needed.
    graph.update(dict((item, set()) for item in extra_items_in_deps))

    return graph


def get_subgraphs(graph=None):
    """
    Given a graph of possibly disconnected components, generate all graphs of
    connected components. graph is a dictionary of dependencies. Keys are
    components, and values are sets of components on which they depend.
    """
    graph = graph or DEPENDENCIES
    keys = set(graph)
    frontier = set()
    seen = set()
    while keys:
        frontier.add(keys.pop())
        while frontier:
            component = frontier.pop()
            seen.add(component)
            frontier |= set([d for d in get_dependencies(component) if d in graph])
            frontier |= set([d for d in get_dependents(component) if d in graph])
            frontier -= seen
        yield dict((s, get_dependencies(s)) for s in seen)
        keys -= seen
        seen.clear()


def _import(path, continue_on_error):
    log.debug("Importing %s" % path)
    try:
        return importlib.import_module(path)
    except Exception as ex:
        log.exception(ex)
        if not continue_on_error:
            raise


def _load_components(path, include=".*", exclude="test", continue_on_error=True):
    num_loaded = 0
    if path.endswith(".py"):
        path, _ = os.path.splitext(path)

    path = path.rstrip("/").replace("/", ".")

    package = _import(path, continue_on_error)
    if not package:
        return 0

    num_loaded += 1

    do_include = re.compile(include).search if include else lambda x: True
    do_exclude = re.compile(exclude).search if exclude else lambda x: False

    if not hasattr(package, "__path__"):
        return num_loaded

    prefix = package.__name__ + "."
    for _, name, is_pkg in pkgutil.iter_modules(path=package.__path__, prefix=prefix):
        if not name.startswith(prefix):
            name = prefix + name
        if is_pkg:
            num_loaded += _load_components(name, include, exclude, continue_on_error)
        else:
            if do_include(name) and not do_exclude(name):
                _import(name, continue_on_error)
                num_loaded += 1

    return num_loaded


def load_components(*paths, **kwargs):
    """
    Loads all components on the paths. Each path should be a package or module.
    All components beneath a path are loaded.

    Args:
        paths (str): A package or module to load

    Keyword Args:
        include (str): A regular expression of packages and modules to include.
            Defaults to '.*'
        exclude (str): A regular expression of packges and modules to exclude.
            Defaults to 'test'
        continue_on_error (bool): If True, continue importing even if something
            raises an ImportError. If False, raise the first ImportError.

    Returns:
        int: The total number of modules loaded.

    Raises:
        ImportError
    """
    num_loaded = 0
    for path in paths:
        num_loaded += _load_components(path, **kwargs)
    return num_loaded


def first_of(dependencies, broker):
    for d in dependencies:
        if d in broker:
            return broker[d]


def split_requirements(requires):
    req_all = []
    req_any = []
    for r in requires:
        if isinstance(r, list):
            req_any.append(r)
        else:
            req_all.append(r)
    return req_all, req_any


def stringify_requirements(requires):
    if isinstance(requires, tuple):
        req_all, req_any = requires
    else:
        req_all, req_any = split_requirements(requires)
    pretty_all = [get_name(r) for r in req_all]
    pretty_any = [str([get_name(r) for r in any_list]) for any_list in req_any]
    result = "All: %s" % pretty_all + " Any: " + " Any: ".join(pretty_any)
    return result


def _register_component(delegate):
    component = delegate.component

    dependencies = delegate.get_dependencies()
    DEPENDENCIES[component] = dependencies
    COMPONENTS[delegate.group][component] |= dependencies

    COMPONENTS_BY_TYPE[delegate.type].add(component)
    for k, v in COMPONENTS_BY_TYPE.items():
        if issubclass(delegate.type, k) and delegate.type is not k:
            v.add(component)
    DELEGATES[component] = delegate

    MODULE_NAMES[component] = get_module_name(component)
    BASE_MODULE_NAMES[component] = get_base_module_name(component)


class ComponentType(object):
    """
    ComponentType is the base class for all component type decorators.

    For Example:

    .. code-block:: python

       class my_component_type(ComponentType):
           pass

       @my_component_type(SshDConfig, InstalledRpms, [ChkConfig, UnitFiles], optional=[IPTables, IpAddr])
       def my_func(sshd_config, installed_rpms, chk_config, unit_files, ip_tables, ip_addr):
           return installed_rpms.newest("bash")

    Notice that the arguments to ``my_func`` correspond to the dependencies in
    the ``@my_component_type`` and are in the same order.

    When used, a ``my_component_type`` instance is created whose
    ``__init__`` gets passed dependencies and whose ``__call__`` gets
    passed the component to run if dependencies are met.

    Parameters to the decorator have these forms:

    ============  ===============================  ==========================
    Criteria      Example Decorator Arguments      Description
    ============  ===============================  ==========================
    Required      ``SshDConfig, InstalledRpms``    A regular argument
    At Least One  ``[ChkConfig, UnitFiles]``       An argument as a list
    Optional      ``optional=[IPTables, IpAddr]``  A list following optional=
    ============  ===============================  ==========================

    If a parameter is required, the value provided for it is guaranteed not to
    be ``None``. In the example above, ``sshd_config`` and ``installed_rpms``
    will not be ``None``.

    At least one of the arguments to parameters of an "at least one"
    list will not be ``None``. In the example, either or both of ``chk_config``
    and unit_files will not be ``None``.

    Any or all arguments for optional parameters may be ``None``.

    The following keyword arguments may be passed to the decorator:

    Attributes:
        requires (list): a list of components that all components decorated with
            this type will implicitly require. Additional components passed to
            the decorator will be appended to this list.
        optional (list): a list of components that all components decorated with
            this type will implicitly depend on optionally. Additional components
            passed as ``optional`` to the decorator will be appended to this list.
        metadata (dict): an arbitrary dictionary of information to associate
            with the component you're decorating. It can be retrieved with
            ``get_metadata``.
        tags (list): a list of strings that categorize the component. Useful for
            formatting output or sifting through results for components you care
            about.
        group: ``GROUPS.single`` or ``GROUPS.cluster``. Used to organize
            components into "groups" that run together with :func:`insights.core.dr.run`.
        cluster (bool): if ``True`` will put the component into the
            ``GROUPS.cluster`` group. Defaults to ``False``. Overrides ``group``
            if ``True``.

    """

    requires = []
    """
    a list of components that all components decorated with this type will
    implicitly require. Additional components passed to the decorator will be
    appended to this list.
    """
    optional = []
    """
    a list of components that all components decorated with this type will
    implicitly depend on optionally. Additional components passed as
    ``optional`` to the decorator will be appended to this list.
    """
    metadata = {}
    """
    an arbitrary dictionary of information to associate with the component
    you're decorating. It can be retrieved with ``get_metadata``.
    """
    tags = []
    """
    a list of strings that categorize the component. Useful for formatting
    output or sifting through results for components you care about.
    """
    group = GROUPS.single
    """
    group: ``GROUPS.single`` or ``GROUPS.cluster``. Used to organize components
    into "groups" that run together with :func:`insights.core.dr.run`.
    """

    def __init__(self, *deps, **kwargs):
        """
        This constructor is the parameterized part of a decorator.
        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.component = None
        self.requires = []
        self.at_least_one = []
        self.deps = []
        self.type = self.__class__

        deps = list(deps) or kwargs.get("requires", [])

        requires = list(self.__class__.requires) + deps
        for d in requires:
            if isinstance(d, list):
                self.at_least_one.append(d)
                self.deps.extend(d)
            else:
                self.requires.append(d)
                self.deps.append(d)

        self.optional = list(self.__class__.optional)
        optional = kwargs.get("optional", [])
        if optional and not isinstance(optional, list):
            optional = [optional]
        self.optional.extend(optional)

        self.deps.extend(self.optional)

        self.dependencies = set(self.deps)

        self.metadata = {}
        self.metadata.update(self.__class__.metadata)
        self.metadata.update(kwargs.get("metadata", {}) or {})

        self.group = kwargs.get("group", self.__class__.group)
        if kwargs.get("cluster", False):
            self.group = GROUPS.cluster

        tags = []
        tags.extend(self.__class__.tags)
        tags.extend(kwargs.get("tags", []) or [])
        self.tags = set(tags)

    def __call__(self, component):
        """
        This function is the part of the decorator that receives the function
        or class.
        """
        self.component = component
        self.__name__ = component.__name__
        self.__module__ = component.__module__
        self.__doc__ = component.__doc__
        self.__qualname__ = getattr(component, "__qualname__", None)
        for d in self.dependencies:
            add_dependent(d, component)
        _register_component(self)
        return component

    def invoke(self, results):
        """
        Handles invocation of the component. The default implementation invokes
        it with positional arguments based on order of dependency declaration.
        """
        args = [results.get(d) for d in self.deps]
        return self.component(*args)

    def get_missing_dependencies(self, broker):
        """
        Gets required and at-least-one dependencies not provided by the broker.
        """
        missing_required = [r for r in self.requires if r not in broker]
        missing_at_least_one = [d for d in self.at_least_one if not set(d).intersection(broker)]
        if missing_required or missing_at_least_one:
            return (missing_required, missing_at_least_one)

    def process(self, broker):
        """
        Ensures dependencies have been met before delegating to `self.invoke`.
        """
        if any(i in broker for i in IGNORE.get(self.component, [])):
            raise SkipComponent()
        missing = self.get_missing_dependencies(broker)
        if missing:
            raise MissingRequirements(missing)
        return self.invoke(broker)

    def get_dependencies(self):
        return self.dependencies

    def add_dependency(self, dep):
        group = self.group
        self.at_least_one[0].append(dep)
        self.deps.append(dep)
        self.dependencies.add(dep)
        add_dependent(dep, self.component)

        DEPENDENCIES[self.component].add(dep)
        COMPONENTS[group][self.component].add(dep)


class Broker(object):
    """
    The Broker is a fancy dictionary that keeps up with component instances as
    a graph is evaluated. It's the state of the evaluation. Once a graph has
    executed, the broker will contain everything about the evaluation:
    component instances, timings, exceptions, and tracebacks.

    You can either inspect the broker at the end of an evaluation, or you can
    register callbacks with it, and they'll get invoked after each component
    is called.

    Attributes:
        instances (dict): the component instances with components as keys.
        missing_requirements (dict): components that didn't have their dependencies
            met. Values are a two-tuple. The first element is the list of
            required dependencies that were missing. The second element is the
            list of "at least one" dependencies that were missing. For more
            information on dependency types, see the :class:`ComponentType`
            docs.
        exceptions (defaultdict(list)): Components that raise any type of
            exception except :class:`SkipComponent` during evaluation. The key
            is the component, and the value is a list of exceptions. It's a
            list because some components produce multiple instances.
        tracebacks (dict): keys are exceptions and values are their text
            tracebacks.
        exec_times (dict): component -> float dictionary where values are the
            number of seconds the component took to execute. Calculated using
            :func:`time.time`. For components that produce multiple instances,
            the execution time here is the sum of their individual execution
            times.
    """
    def __init__(self, seed_broker=None):
        self.instances = dict(seed_broker.instances) if seed_broker else {}
        self.missing_requirements = {}
        self.exceptions = defaultdict(list)
        self.tracebacks = {}
        self.exec_times = {}

        self.observers = defaultdict(set)
        if seed_broker is not None:
            self.observers.update(seed_broker.observers)
        else:
            self.observers[ComponentType] = set()
            for k, v in TYPE_OBSERVERS.items():
                self.observers[k] |= set(v)

    def observer(self, component_type=ComponentType):
        """
        You can use ``@broker.observer()`` as a decorator to your callback
        instead of :func:`Broker.add_observer`.
        """
        def inner(func):
            self.add_observer(func, component_type)
            return func
        return inner

    def add_observer(self, o, component_type=ComponentType):
        """
        Add a callback that will get invoked after each component is called.

        Args:
            o (func): the callback function

        Keyword Args:
            component_type (ComponentType): the :class:`ComponentType` to observe.
                The callback will fire any time an instance of the class or its
                subclasses is invoked.
        The callback should look like this:

        .. code-block:: python

            def callback(comp, broker):
                value = broker.get(comp)
                # do something with value
                pass

        """

        self.observers[component_type].add(o)

    def fire_observers(self, component):
        _type = get_component_type(component)
        if not _type:
            return

        for k, v in self.observers.items():
            if issubclass(_type, k):
                for o in v:
                    try:
                        o(component, self)
                    except Exception as e:
                        log.exception(e)

    def add_exception(self, component, ex, tb=None):
        if isinstance(ex, MissingRequirements):
            self.missing_requirements[component] = ex.requirements
        else:
            self.exceptions[component].append(ex)
            self.tracebacks[ex] = tb

    def __iter__(self):
        return iter(self.instances)

    def keys(self):
        return self.instances.keys()

    def items(self):
        return self.instances.items()

    def values(self):
        return self.instances.values()

    def get_by_type(self, _type):
        """
        Return all of the instances of :class:`ComponentType` ``_type``.
        """
        r = {}
        for k, v in self.items():
            if get_component_type(k) is _type:
                r[k] = v
        return r

    def __contains__(self, component):
        return component in self.instances

    def __setitem__(self, component, instance):
        msg = "Already exists in broker with key: %s"
        if component in self.instances:
            raise KeyError(msg % get_name(component))

        self.instances[component] = instance

    def __delitem__(self, component):
        if component in self.instances:
            del self.instances[component]
            return

    def __getitem__(self, component):
        if component in self.instances:
            return self.instances[component]

        raise KeyError("Unknown component: %s" % get_name(component))

    def get(self, component, default=None):
        try:
            return self[component]
        except KeyError:
            return default

    def print_component(self, component_type):
        print(json.dumps(
            dict((get_name(c), self[c])
                 for c in sorted(self.get_by_type(component_type), key=get_name))))


def get_missing_requirements(func, requires, d):
    """
    .. deprecated:: 1.x
    """
    if not requires:
        return None
    if any(i in d for i in IGNORE.get(func, [])):
        raise SkipComponent()
    req_all, req_any = split_requirements(requires)
    d = set(d.keys())
    req_all = [r for r in req_all if r not in d]
    req_any = [r for r in req_any if set(r).isdisjoint(d)]
    if req_all or req_any:
        return req_all, req_any
    else:
        return None


def add_observer(o, component_type=ComponentType):
    """
    Add a callback that will get invoked after each component is called.

    Args:
        o (func): the callback function

    Keyword Args:
        component_type (ComponentType): the :class:`ComponentType` to observe.
            The callback will fire any time an instance of the class or its
            subclasses is invoked.

    The callback should look like this:

    .. code-block:: python

        def callback(comp, broker):
            value = broker.get(comp)
            # do something with value
            pass

    """

    TYPE_OBSERVERS[component_type].add(o)


def observer(component_type=ComponentType):
    """
    You can use ``@broker.observer()`` as a decorator to your callback
    instead of :func:`add_observer`.
    """
    def inner(func):
        add_observer(func, component_type)
        return func
    return inner


def run_order(graph):
    """
    Returns components in an order that satisfies their dependency
    relationships.
    """
    return toposort_flatten(graph, sort=False)


def _determine_components(components):
    if isinstance(components, dict):
        return components

    if hashable(components) and components in COMPONENTS_BY_TYPE:
        components = get_components_of_type(components)

    if isinstance(components, (list, set)):
        graph = {}
        for c in components:
            graph.update(get_dependency_graph(c))
        return graph

    if hashable(components) and components in DELEGATES:
        return get_dependency_graph(components)

    if hashable(components) and components in COMPONENTS:
        return COMPONENTS[components]


def run(components=None, broker=None):
    """
    Executes components in an order that satisfies their dependency
    relationships.

    Keyword Args:
        components: Can be one of a dependency graph, a single component, a
            component group, or a component type. If it's anything other than a
            dependency graph, the appropriate graph is built for you and before
            evaluation.
        broker (Broker): Optionally pass a broker to use for evaluation. One is
            created by default, but it's often useful to seed a broker with an
            initial dependency.
    Returns:
        Broker: The broker after evaluation.
    """
    components = components or COMPONENTS[GROUPS.single]
    components = _determine_components(components)
    broker = broker or Broker()

    for component in run_order(components):
        start = time.time()
        try:
            if (component not in broker and component in components and
               component in DELEGATES and
               is_enabled(component)):
                log.info("Trying %s" % get_name(component))
                result = DELEGATES[component].process(broker)
                broker[component] = result
        except MissingRequirements as mr:
            if log.isEnabledFor(logging.DEBUG):
                name = get_name(component)
                reqs = stringify_requirements(mr.requirements)
                log.debug("%s missing requirements %s" % (name, reqs))
            broker.add_exception(component, mr)
        except SkipComponent:
            pass
        except Exception as ex:
            tb = traceback.format_exc()
            log.warn(tb)
            broker.add_exception(component, ex, tb)
        finally:
            broker.exec_times[component] = time.time() - start
            broker.fire_observers(component)

    return broker


def generate_incremental(components=None, broker=None):
    components = components or COMPONENTS[GROUPS.single]
    components = _determine_components(components)
    seed_broker = broker or Broker()
    for graph in get_subgraphs(components):
        broker = Broker(seed_broker)
        yield (graph, broker)


def run_incremental(components=None, broker=None):
    """
    Executes components in an order that satisfies their dependency
    relationships. Disjoint subgraphs are executed one at a time and a broker
    containing the results for each is yielded. If a broker is passed here, its
    instances are used to seed the broker used to hold state for each sub
    graph.

    Keyword Args:
        components: Can be one of a dependency graph, a single component, a
            component group, or a component type. If it's anything other than a
            dependency graph, the appropriate graph is built for you and before
            evaluation.
        broker (Broker): Optionally pass a broker to use for evaluation. One is
            created by default, but it's often useful to seed a broker with an
            initial dependency.
    Yields:
        Broker: the broker used to evaluate each subgraph.
    """
    for graph, _broker in generate_incremental(components, broker):
        yield run(graph, broker=_broker)


def run_all(components=None, broker=None, pool=None):
    if pool:
        futures = []
        for graph, _broker in generate_incremental(components, broker):
            futures.append(pool.submit(run, graph, _broker))
        return [f.result() for f in futures]
    else:
        return list(run_incremental(components=components, broker=broker))



from .. import Parser, parser, get_active_lines, LegacyItemAccess
from insights.specs import Specs


@parser(Specs.yum_repos_d)
class YumReposD(LegacyItemAccess, Parser):
    """Class to parse the files under ``yum.repos.d`` """

    def get(self, key):
        return self.data.get(key)

    def parse_content(self, content):
        '''
        Return an object contains a dict.
        {
            "rhel-source": {
                "gpgcheck": "1",
                "gpgkey": ["file:///etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-release",
                           "file:///etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-release_bak"]
                "enabled": "0",
                "name": "Red Hat Enterprise Linux $releasever - $basearch - Source",
                "baseurl": "ftp://ftp.redhat.com/pub/redhat/linux/enterprise/$releasever/en/os/SRPMS/"
            }
        }
        ----------------------------------------------------
        There are several files in 'yum.repos.d' directory, which have the same
        format.  For example:
        --------one of the files : rhel-source.repo---------
        [rhel-source]
        name=Red Hat Enterprise Linux $releasever - $basearch - Source
        baseurl=ftp://ftp.redhat.com/pub/redhat/linux/enterprise/$releasever/en/os/SRPMS/
        enabled=0
        gpgcheck=1
        gpgkey=file:///etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-release
               file:///etc/pki/rpm-gpg/RPM-GPG-KEY-redhat-release_bak
        '''
        repos_dict = {}
        section_dict = {}
        key = None
        for line in get_active_lines(content):
            if line.startswith('['):
                section_dict = {}
                repos_dict[line[1:-1]] = section_dict
            elif '=' in line:
                key, value = [s.strip() for s in line.split("=", 1)]
                if key in ('baseurl', 'gpgkey'):
                    section_dict[key] = [v.strip() for v in value.split(",")]
                else:
                    section_dict[key] = value
            else:
                if key and isinstance(section_dict[key], list):
                    section_dict[key].extend(v.strip() for v in line.split(","))
                # Otherwise ignore line if no key or we don't store multiple values
        self.data = repos_dict

    def __iter__(self):
        for repo in self.data:
            yield repo



"""
Systemd command ``show`` outputs
================================
This command shows all the paramters for this service, these parameters can be set
through unit files corresponding to that service or from command line.

Parsers included in this module are:

SystemctlShowCinderVolume - command ``systemctl show openstack-cinder-volume``
------------------------------------------------------------------------------

SystemctlShowMariaDB - command ``systemctl show mariadb``
---------------------------------------------------------

SystemctlShowPulpWorkers - command ``systemctl show pulp_workers``
------------------------------------------------------------------

SystemctlShowPulpResourceManager - command ``systemctl show pulp_resource_manager``
-----------------------------------------------------------------------------------

SystemctlShowPulpCelerybeat - command ``systemctl show pulp_celerybeat``
------------------------------------------------------------------------

SystemctlShowHttpd - command ``systemctl show httpd``
-----------------------------------------------------

SystemctlShowQpidd - command ``systemctl show qpidd``
-----------------------------------------------------

SystemctlShowQdrouterd - command ``systemctl show qdrouterd``
-------------------------------------------------------------

SystemctlShowSmartpdc - command ``systemctl show smart_proxy_dynflow_core``
---------------------------------------------------------------------------
"""


from .. import LegacyItemAccess
from .. import CommandParser
from .. import parser
from ..parsers import split_kv_pairs
from ..specs import Specs


class SystemctlShow(LegacyItemAccess, CommandParser):
    """Class for parsing ``systemctl show <Service_Name>`` command output"""

    def parse_content(self, content):
        """
        Sample Input::

            TimeoutStartUSec=1min 30s
            LimitNOFILE=65536
            LimitMEMLOCK=
            LimitLOCKS=18446744073709551615

        Sample Output::

            {"LimitNOFILE"     : "65536",
            "TimeoutStartUSec" : "1min 30s",
            "LimitLOCKS"       : "18446744073709551615"}

        In CMD's output, empty properties are suppressed by default.
        We will also suppressed empty properties in return data.
        """
        data = {}
        data = split_kv_pairs(content, use_partition=False)

        """Remove empty key"""
        self.data = dict((k, v) for k, v in data.items() if not v == '')


@parser(Specs.systemctl_cinder_volume)
class SystemctlShowCinderVolume(SystemctlShow):
    """
    Class for ``systemctl show openstack-cinder-volume``.

    Typical output of ``/bin/systemctl show openstack-cinder-volume`` command is::

        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        WatchdogUSec=0
        LimitCORE=18446744073709551615
        LimitRSS=18446744073709551615
        LimitNOFILE=65536
        LimitAS=18446744073709551615
        LimitNPROC=63391
        Transient=no
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_cinder_volume["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_mariadb)
class SystemctlShowMariaDB(SystemctlShow):
    """
    Class for ``systemctl show mariadb``.

    Typical output of ``/bin/systemctl show mariadb`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStopUSec=5min
        ExecStartPre={ path=/usr/libexec/mariadb-prepare-db-dir ; argv[]=/usr/libexec/mariadb-prepare-db-dir %n ; ignore_errors=no ; start_time=[Mon 2017-05-22 06:49:01 EDT] ; stop_time=[Mon 2017-05-22 06:49:02 EDT] ; pid=1946 ; code=exited ; status=0 }
        ExecStart={ path=/usr/bin/mysqld_safe ; argv[]=/usr/bin/mysqld_safe --basedir=/usr ; ignore_errors=no ; start_time=[Mon 2017-05-22 06:49:02 EDT] ; stop_time=[n/a] ; pid=2304 ; code=(null) ; status=0/0 }
        ExecStartPost={ path=/usr/libexec/mariadb-wait-ready ; argv[]=/usr/libexec/mariadb-wait-ready $MAINPID ; ignore_errors=no ; start_time=[Mon 2017-05-22 06:49:02 EDT] ; stop_time=[Mon 2017-05-22 06:49:12 EDT] ; pid=2305 ; code=exited ; status=0 }
        Slice=system.slice
        ControlGroup=/system.slice/mariadb.service
        After=network.target -.mount systemd-journald.socket tmp.mount basic.target syslog.target system.slice
        MemoryCurrent=18446744073709551615
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_mariadb["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_pulp_workers)
class SystemctlShowPulpWorkers(SystemctlShow):
    """
    Class for ``systemctl show pulp_workers``.

    Typical output of ``/bin/systemctl show pulp_workers`` command is::

        Type=oneshot
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=0
        TimeoutStopUSec=1min 30s
        WatchdogUSec=0
        WatchdogTimestampMonotonic=0
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:33 CST
        ExecMainStartTimestampMonotonic=105521850
        ExecMainExitTimestamp=Thu 2018-01-11 14:22:33 CST
        ExecMainExitTimestampMonotonic=105593405
        ExecStart={ path=/usr/libexec/pulp-manage-workers ; argv[]=/usr/libexec/pulp-manage-workers start ; ignore_err
        ExecStop={ path=/usr/libexec/pulp-manage-workers ; argv[]=/usr/libexec/pulp-manage-workers stop ; ignore_error
        Slice=system.slice
        After=systemd-journald.socket system.slice network.target basic.target
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_pulp_workers["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_pulp_resmg)
class SystemctlShowPulpResourceManager(SystemctlShow):
    """
    Class for ``systemctl show pulp_resource_manager``.

    Typical output of ``/bin/systemctl show pulp_resource_manager`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:33 CST
        ExecMainStartTimestampMonotonic=105028117
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2810
        ExecMainCode=0
        ExecMainStatus=0
        ExecStart={ path=/usr/bin/celery ; argv[]=/usr/bin/celery worker -A pulp.server.async.app -n resource_manager@
        Slice=system.slice
        After=basic.target network.target system.slice -.mount systemd-journald.socket
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_pulp_resource_manager["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_pulp_celerybeat)
class SystemctlShowPulpCelerybeat(SystemctlShow):
    """
    Class for ``systemctl show pulp_celerybeat``.

    Typical output of ``/bin/systemctl show pulp_celerybeat`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:32 CST
        ExecMainStartTimestampMonotonic=104261679
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2747
        ExecMainCode=0
        ExecMainStatus=0
        ExecStart={ path=/usr/bin/celery ; argv[]=/usr/bin/celery beat --scheduler=pulp.server.async.scheduler.Schedul
        Slice=system.slice
        After=basic.target network.target system.slice -.mount systemd-journald.socket
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_pulp_celerybeat["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_httpd)
class SystemctlShowHttpd(SystemctlShow):
    """
    Class for ``systemctl show httpd``.

    Typical output of ``/bin/systemctl show httpd`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:32 CST
        ExecMainStartTimestampMonotonic=104261679
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2747
        ExecMainCode=0
        ExecMainStatus=0
        ExecStart={ path=/usr/sbin/httpd ; argv[]=/usr/sbin/httpd $OPTIONS -DFOREGROUND ; ignore_errors=no ; start_time=[Tue 2018-05-15 09:30:08 CST] ; stop_time=[n/a] ; pid=1605 ; code=(null) ; status=0/0 }
        ExecReload={ path=/usr/sbin/httpd ; argv[]=/usr/sbin/httpd $OPTIONS -k graceful ; ignore_errors=no ; start_time=[Wed 2018-05-16 03:07:01 CST] ; stop_time=[Wed 2018-05-16 03:07:01 CST] ; pid=21501 ; code=exited ; status=0 }
        ExecStop={ path=/bin/kill ; argv[]=/bin/kill -WINCH ${MAINPID} ; ignore_errors=no ; start_time=[n/a] ; stop_time=[n/a] ; pid=0 ; code=(null) ; status=0/0 }
        Slice=system.slice
        ControlGroup=/system.slice/httpd.service
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_httpd["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_qpidd)
class SystemctlShowQpidd(SystemctlShow):
    """
    Class for ``systemctl show qpidd``.

    Typical output of ``/bin/systemctl show qpidd`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:32 CST
        ExecMainStartTimestampMonotonic=104261679
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2747
        ExecMainCode=0
        ExecMainStatus=0
        ExecStart={ path=/usr/sbin/qpidd ; argv[]=/usr/sbin/qpidd --config /etc/qpid/qpi
        Slice=system.slice
        ControlGroup=/system.slice/qpidd.service
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_qpidd["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_qdrouterd)
class SystemctlShowQdrouterd(SystemctlShow):
    """
    Class for ``systemctl show qdrouterd``.

    Typical output of ``/bin/systemctl show qdrouterd`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:32 CST
        ExecMainStartTimestampMonotonic=104261679
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2747
        ExecMainCode=0
        ExecMainStatus=0
        Slice=system.slice
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_qdrouterd["LimitNOFILE"]
        '4096'

    """
    pass


@parser(Specs.systemctl_smartpdc)
class SystemctlShowSmartpdc(SystemctlShow):
    """
    Class for ``systemctl show smart_proxy_dynflow_core``.

    Typical output of ``/bin/systemctl show smart_proxy_dynflow_core`` command is::

        Type=simple
        Restart=no
        NotifyAccess=none
        RestartUSec=100ms
        TimeoutStartUSec=1min 30s
        TimeoutStopUSec=1min 30s
        ExecMainStartTimestamp=Thu 2018-01-11 14:22:32 CST
        ExecMainStartTimestampMonotonic=104261679
        ExecMainExitTimestampMonotonic=0
        ExecMainPID=2747
        ExecMainCode=0
        ExecMainStatus=0
        Slice=system.slice
        LimitNOFILE=4096
        ...

    Examples:
        >>> systemctl_show_smartpdc["LimitNOFILE"]
        '4096'

    """
    pass



import doctest

from insights.parsers import swift_log
from insights.tests import context_wrap

SWIFT_LOG_CONTENT = """
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter listing_formats to pipeline at position 4
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter gatekeeper to pipeline at position 1
Mar  1 13:53:47 u1lab5 proxy-server: Pipeline was modified. New pipeline is "catch_errors gatekeeper healthcheck proxy-logging cache listing_formats ratelimit bulk tempurl formpost authtoken keystone staticweb copy slo dlo versioned_writes proxy-logging proxy-server".
Mar  1 13:53:47 u1lab5 proxy-server: object_post_as_copy=true is deprecated; This option is now ignored
Mar  1 13:53:47 u1lab5 proxy-server: Starting Keystone auth_token middleware
Mar  1 13:53:47 u1lab5 proxy-server: Started child 25651
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter listing_formats to pipeline at position 4
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter gatekeeper to pipeline at position 1
Mar  1 13:53:47 u1lab5 proxy-server: Pipeline was modified. New pipeline is "catch_errors gatekeeper healthcheck proxy-logging cache listing_formats ratelimit bulk tempurl formpost authtoken keystone staticweb copy slo dlo versioned_writes proxy-logging proxy-server".
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter listing_formats to pipeline at position 4
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter gatekeeper to pipeline at position 1
Mar  1 13:53:47 u1lab5 proxy-server: Adding required filter listing_formats to pipeline at position 4
Mar  1 13:53:47 u1lab5 proxy-server: Pipeline was modified. New pipeline is "catch_errors gatekeeper healthcheck proxy-logging cache listing_formats ratelimit bulk tempurl formpost authtoken keystone staticweb copy slo dlo versioned_writes proxy-logging proxy-server".
Mar  1 14:49:13 u1lab5 container-server: 10.40.28.2 - - [01/Mar/2019:14:49:13 +0000] "DELETE /1/177/.expiring_objects/1551398391" 409 - "DELETE http://localhost/v1/.expiring_objects/1551398391" "txdd303f0a04264ca9b9921-005c794669" "proxy-server 25600" 0.0009 "-" 27659 -
Oct  2 09:10:43 u1lab5 object-expirer: STDERR: ERROR:root:Error talking to memcached: [Errno 32] Broken pipe (txn: tx015dd5a60bfa491d8eb09-005bb36e53)
""".strip()

SAMPLE_LOG = """
Sep 29 23:50:29 rh-server object-server: Starting object replication pass.
Sep 29 23:50:29 rh-server object-server: Nothing replicated for 0.01691198349 seconds.
Sep 29 23:50:29 rh-server object-server: Object replication complete. (0.00 minutes)
Sep 29 23:50:38 rh-server container-server: Beginning replication run
Sep 29 23:50:38 rh-server container-server: Replication run OVER
Sep 29 23:50:38 rh-server container-server: Attempted to replicate 0 dbs in 0.00064 seconds (0.00000/s)
""".strip()


def test_swift_log():
    log = swift_log.SwiftLog(context_wrap(SWIFT_LOG_CONTENT))
    assert "Starting Keystone auth_token middleware" in log
    obj_expirer_lines = log.get("object-expirer")
    assert len(obj_expirer_lines) == 1
    assert obj_expirer_lines[0].get("procname") == "object-expirer"
    assert obj_expirer_lines[0].get("timestamp") == "Oct 2 09:10:43"
    assert obj_expirer_lines[0].get("message") == "STDERR: ERROR:root:Error talking to memcached: [Errno 32] Broken pipe (txn: tx015dd5a60bfa491d8eb09-005bb36e53)"

    log_line = 'Sep 29 23:50:55 pbnec2-l-rh-ocld-2 object-server: Begin object audit "forever" mode (ALL)'
    log = swift_log.SwiftLog(context_wrap(log_line))
    assert "Begin object audit" in log
    assert log.get("Begin object audit")[0]['raw_message'] == log_line


def test_documentation():
    failed_count, tests = doctest.testmod(
        swift_log,
        globs={'swift_log': swift_log.SwiftLog(context_wrap(SAMPLE_LOG))}
    )
    assert failed_count == 0



"""
configtree models configurations that are similar to dictionaries except that
each option may have a list of un-named attributes, and duplicate options are
allowed. Compound options containing children in addition to attributes also
are supported.

This model fits many systems that have a main configuration file with
supplementary files included in it by reference. Examples are httpd and nginx,
although simple .ini files are also supported. Base classes for Insights
parsers and combiners handle incorporating master and child documents into a
single tree. They make integration with Insights concise.

Individual document parsers are responsible for transforming raw content into
the configtree model. Some base classes and primitive parsing functions are
provided, but most systems' files (except maybe .ini) need specific handling.

configtree provides a small DSL for querying its trees, so navigating them for
specific information is less tedious and error prone.

Generating documents of various formats from a master tree is straightforward.
"""
import operator
import re
import six
from functools import partial
from itertools import chain


# classes modeling configuration trees
class Node(object):
    def __init__(self, name=None, attrs=None, children=None, ctx=None):
        """
        Base object of the document model. The documents are similar to
        dictionaries except that each level may have a list of un-named
        attributes, and duplicate keys at each level may be allowed.

        Attributes:
            name (str): something like "Directory" or "ServerRoot" from
                httpd.conf, "http" from nginx, or the name of an option in an
                ini file.
            attrs (list): the list of attributes associated with the name.
            children (list): child Nodes of this Node.
            ctx (ConfigParser):  provides lineat function for accessing the raw
                text from which the node was parsed.
        """
        self.name = name
        self.attrs = attrs or []
        self.children = children or []
        self.ctx = ctx
        self.pos = None
        self.parent = None

    def select(self, *queries, **kwargs):
        """
        Given a list of queries, executes those queries against the set of
        Nodes. A Node has three primary attributes: name (str), attrs
        ([str|int]), and children ([Node]).

        Nodes also have a value attribute that is either the first attribute
        (in the case of simple directives that only have one), or the string
        representation of all attributes joined by a single space.

        Each positional argument to select represents a query against the name
        and/or attributes of the corresponding level of the configuration tree.
        The first argument queries root nodes, the second argument queries
        children of the root nodes, etc.

        An individual query is either a single value or a tuple. A single value
        queries the name of a Node. A tuple queries the name and the attrs.

        So: `select(name_predicate)` or `select((name_predicate,
        attrs_predicate))`

        In general, `select(pred1, pred2, pred3, ...)`

        If a predicate is a simple value (string or int), an exact match is
        required for names, and an exact match of any attribute is required for
        attributes.

        Examples:
        `select("Directory")` queries for all root nodes named Directory.

        `select("Directory", "Options")` queries for all root nodes named
        Directory that contain at least one child node named Options. Notice
        the argument positions: Directory is in position 1, and Options is in
        position 2.

        `select(("Directory", "/"))` queries for all root nodes named Directory
        that contain an attribute exactly matching "/". Notice this is one
        argument to select: a 2-tuple with predicates for name and attrs.

        If you are only interested in attributes, just pass `None` for the name
        predicate in the tuple: `select((None, "/"))` will return all root
        nodes with at least one attribute of "/"

        In addition to exact matches, the elements of a query can be functions
        that accept the value corresponding to their position in the query. A
        handful of useful functions and boolean operators between them are
        provided.

        `select(startswith("Dir"))` queries for all root nodes with names
        starting with "Dir".

        `select(~startswith("Dir"))` queries for all root nodes with names not
        starting with "Dir".

        `select(startswith("Dir") | startswith("Ali"))` queries for all root
        nodes with names starting with "Dir" or "Ali". The return of `|` is a
        single callable passed in the first argument position of select.

        `select(~startswith("Dir") & ~startswith("Ali"))` queries for all root
        nodes with names not starting with "Dir" or "Ali".

        If a function is in an attribute position, it is considered True if it
        returns True for any attribute.

        For example, `select((None, 80))` often will return the list of one
        Node [Listen 80]

        `select(("Directory", startswith("/var")))` will return all root nodes
        named Directory that also have an attribute starting with "/var"

        If you know that your selection will only return one element, or you
        only want the first or last result of the query , pass `one=first` or
        `one=last`.

        `select(("Directory", startswith("/")), one=last)` will return the
        single root node for the last Directory entry starting with "/"

        If instead of the root nodes that match you want the child nodes that
        caused the match, pass `roots=False`.

        `node = select(("Directory", "/var/www/html"), "Options", one=last,
        roots=False)` might return the Options node if the Directory for
        "/var/www/html" was defined and contained an Options Directive. You
        could then access the attributes with `node.attrs`. If the query didn't
        match anything, it would have returned None.

        If you want to slide the query down the branches of the config, pass
        deep=True to select.  That allows you to do `conf.select("Directory",
        deep=True, roots=False)` and get back all Directory nodes regardless of
        nesting depth.

        conf.select() returns everything.

        Available predicates are:
        & (infix boolean and)
        | (infix boolean or)
        ~ (prefix boolean not)

        For ints or strings:
        eq (==)  e.g. conf.select("Directory, ("StartServers", eq(4)))
        ge (>=)  e.g. conf.select("Directory, ("StartServers", ge(4)))
        gt (>)
        le (<=)
        lt (<)

        For strings:
        contains
        endswith
        startswith
        """
        return select(*queries, **kwargs)(self.children)

    def find(self, *queries, **kwargs):
        """
        Finds the first result found anywhere in the configuration. Pass
        `one=last` for the last result. Returns `None` if no results are found.
        """
        kwargs["deep"] = True
        kwargs["roots"] = False
        if "one" not in kwargs:
            kwargs["one"] = first
        return self.select(*queries, **kwargs)

    def find_all(self, *queries):
        """
        Find all results matching the query anywhere in the configuration.
        Returns an empty `SearchResult` if no results are found.
        """
        return self.select(*queries, deep=True, roots=False)

    def _children_of_type(self, t):
        return [c for c in self.children if isinstance(c, t)]

    @property
    def sections(self):
        return SearchResult(children=self._children_of_type(Section))

    @property
    def directives(self):
        return SearchResult(children=self._children_of_type(Directive))

    def __getitem__(self, query):
        """
        Similar to select, except tuples are constructed without parentheses:
        `conf["Directory", "/"]` is the same as `conf.select(("Directory", "/"))`

        Notice that queries return `SearchResult` instances, which also have
        `__getitem__` delegating to `select` except the select is against
        grandchild nodes in the tree. This allows more complicated queries,
        like this one that gets all Options entries beneath the Directory
        entries starting with "/var":

        `conf["Directory", startswith("/var")]["Options"]`

        or equivalently

        `conf.select(("Directory", startswith("/var")), "Options, roots=False)

        Note you can recover the enclosing section with `node.parent` or a
        node's ultimate root with `node.root`. If a `Node` is a root,
        `node.root` is just the node itself.

        To get all root level Directory and Alias directives, you could do
        something like `conf[eq("Directory") | eq("Alias")]`

        To get loaded auth modules:
        `conf["LoadModule", contains("auth")]`
        """
        if isinstance(query, (int, slice)):
            return self.children[query]
        return select(query, roots=False)(self.children)

    @property
    def root(self):
        p = self
        while p and p.parent:
            p = p.parent
        return p

    @property
    def grandchildren(self):
        return list(chain.from_iterable(c.children for c in self.children))

    @property
    def section(self):
        return None

    @property
    def section_name(self):
        return None

    @property
    def line(self):
        if self.ctx and self.pos is not None:
            return self.ctx.lineat(self.pos)

    @property
    def lineno(self):
        if self.pos is not None:
            return self.pos + 1

    @property
    def file_path(self):
        if self.ctx:
            return self.ctx.file_path

    @property
    def file_name(self):
        if self.ctx:
            return self.ctx.file_name

    @property
    def value(self):
        if len(self.attrs) == 1:
            return self.attrs[0]
        return self.svalue()

    def svalue(self, conv=str):
        return " ".join([conv(a) for a in self.attrs])

    def __bool__(self):
        return bool(self.name or self.attrs or self.children)

    def __contains__(self, item):
        return any(c.name == item for c in self.children)

    def __len__(self):
        return len(self.children)

    def __iter__(self):
        return iter(self.children)

    def __repr__(self):
        return "\n".join(pretty_format(self))

    def __str__(self):
        return self.__repr__()

    __nonzero__ = __bool__


class SearchResult(Node):
    """
    Root node of select results. It has no name or attrs.

    Dictionary operators target grandchildren instead of children
    so that searches can be chained to reach into the tree.
    """
    def __contains__(self, item):
        return any(c.name == item for c in self.grandchildren)

    def __getitem__(self, query):
        if isinstance(query, (int, slice)):
            return self.children[query]
        return select(query, roots=False)(self.grandchildren)


class Root(Node):
    """
    Placeholder Node for top level documents. It has no name or attrs.
    """
    pass


class Section(Node):
    """
    Node for sections such as the XML-like stanzas in httpd configurations.
    """
    def __init__(self, **kwargs):
        super(Section, self).__init__(**kwargs)
        for c in self.children:
            c.parent = self

    @property
    def section(self):
        return self.name

    @property
    def section_name(self):
        return self.value


class Directive(Node):
    """ One liners. A name followed by one or more values. """
    @property
    def section(self):
        if self.parent:
            return self.parent.section

    @property
    def section_name(self):
        if self.parent:
            return self.parent.value


def from_dict(dct):
    """ Convert a dictionary into a configtree.  """
    def inner(d):
        results = []
        for name, v in d.items():
            if isinstance(v, dict):
                results.append(Section(name=name, children=from_dict(v)))
            elif isinstance(v, list):
                if not any(isinstance(i, dict) for i in v):
                    results.append(Directive(name=name, attrs=v))
                else:
                    for i in v:
                        if isinstance(i, dict):
                            results.append(Section(name=name, children=from_dict(i)))
                        elif isinstance(i, list):
                            results.append(Directive(name=name, attrs=i))
                        else:
                            results.append(Directive(name=name, attrs=[i]))
            else:
                results.append(Directive(name, attrs=[v]))
        return results
    return Root(children=inner(dct))


def to_str(x):
    x = str(x)
    if " " in x or "'" in x or '"' in x:
        if "\\'" in x:
            return '"%s"' % x
        if '\\"' in x:
            return "'%s'" % x
        if " " in x:
            return '"%s"' % x
    return x


# pretty formatter for the model
def pretty_format(doc, indent=4):
    results = []

    def inner(d, prepend):
        if isinstance(d, (Root, SearchResult)):
            for c in d.children:
                inner(c, prepend)
        elif isinstance(d, Section):
            results.append("")
            header = d.name if not d.attrs else " ".join([d.name, d.svalue(conv=to_str)])
            results.append(prepend + "[" + header + "]")
            prep = prepend + (" " * indent)
            for c in d.children:
                inner(c, prep)
            results.append("")
        else:
            results.append(prepend + d.name + ": " + d.svalue(conv=to_str))
    inner(doc, "")
    return results


class PushBack(object):
    """
    Wraps an iterable with push back capability. Tracks position in the stream.
    """
    def __init__(self, stream):
        self.stream = iter(stream)
        self.buffer = []
        self.pos = 0

    def push_back(self, item):
        self.pos -= 1
        self.buffer.append(item)

    def peek(self):
        if not self.buffer:
            self.buffer.append(self._next())
        return self.buffer[-1]

    def __iter__(self):
        return self

    def _next(self):
        return next(self.stream)

    def next(self):
        self.pos += 1
        if self.buffer:
            return self.buffer.pop()
        return self._next()

    __next__ = next


class LineGetter(PushBack):
    """
    A push back wrapper for lines. Automatically skips comments and spaces.
    """
    def __init__(self, it, comment_marker="#", strip=True):
        super(LineGetter, self).__init__(it)
        self.comment_marker = comment_marker
        self.strip = strip

    def _next(self):
        l = next(self.stream)
        while l.strip() == "" or l.lstrip().startswith(self.comment_marker):
            self.pos += 1
            l = next(self.stream)
        while l.endswith("\\"):
            self.pos += 1
            l = l.rstrip("\\")
            l += next(self.stream).lstrip()

        return l.strip() if self.strip else l.rstrip()


def eat_whitespace(pb, to=None):
    try:
        while pb.peek().isspace() or pb.peek() == "#":
            if pb.peek().isspace():
                if pb.peek() == to:
                    break
                next(pb)
            if pb.peek() == "#":
                while pb.peek() != "\n":
                    next(pb)
    except StopIteration:
        pass


def parse_string(pb):
    buf = []
    start = next(pb)  # eat quote
    while pb.peek() != start:
        c = next(pb)
        if c == "\\":
            c = next(pb)
        buf.append(c)
    next(pb)  # eat quote
    return "".join(buf)


def parse_bare(pb):
    buf = []
    try:
        while not pb.peek().isspace():
            buf.append(next(pb))
    except StopIteration:
        pass
    return "".join(buf)


BOOLS = {
    "yes": True,
    "on": True,
    "true": True,
    "no": False,
    "off": False,
    "false": False
}


def typed(x):
    v = BOOLS.get(x.lower())
    if v is not None:
        return v
    for t in [int, float]:
        try:
            return t(x)
        except:
            pass
    return x


def parse_attrs(attr_string):
    pb = PushBack(attr_string)
    attrs = []
    while True:
        try:
            while(pb.peek().isspace()):
                next(pb)

            if pb.peek() in ('"', "'"):
                attrs.append(parse_string(pb))
            else:
                attrs.append(parse_bare(pb))
        except StopIteration:
            break
    if len(attrs) == 1:
        return [typed(attrs[0])]
    return attrs


def parse_name_attrs(line, sep=None):
    attrs = []
    parts = re.split(sep, line, 1) if sep else line.split(None, 1)
    name = parts[0].strip()
    if name[0] in ('"', "'"):
        name = parse_string(PushBack(name))
    else:
        name = parse_bare(PushBack(name))
    if len(parts) > 1:
        attrs = parse_attrs(parts[1])
    return (name, attrs)


class DocParser(object):
    """
    Wrapper class so parser functions don't have to thread ctx.

    Other line oriented configuration parsers may subclass this class to
    remove a bit of boilerplate.
    """
    def __init__(self, ctx):
        self.ctx = ctx

    def parse_doc(self, lg):
        children = []
        while True:
            try:
                children.append(self.parse_statement(lg))
            except StopIteration:
                break
        return Root(children=children, ctx=self.ctx)


if six.PY3:
    import unicodedata

    def caseless(text):
        return unicodedata.normalize("NFKD", text.casefold())
else:
    def caseless(text):
        return text.lower()


# DSL for querying trees of Nodes. Start with `select`.
def __or(funcs, args):
    """ Support list sugar for "or" of two predicates. Used inside `select`. """
    results = []
    for f in funcs:
        result = f(args)
        if result:
            results.extend(result)
    return results


# helper functions for Predicates
def _or(f, g):
    def inner(data):
        return f(data) or g(data)
    return inner


def _and(f, g):
    def inner(data):
        return f(data) and g(data)
    return inner


def _not(f):
    def inner(data):
        return not f(data)
    return inner


class Bool(object):
    """ Allows boolean logic between predicates. """
    def __and__(self, other):
        return CompositeBool(_and(self, other))

    def __or__(self, other):
        return CompositeBool(_or(self, other))

    def __invert__(self):
        return CompositeBool(_not(self))


class CompositeBool(Bool):
    """ Combines two DSL predicates. """
    def __init__(self, pred):
        self.pred = pred

    def __call__(self, data):
        try:
            return self.pred(data)
        except:
            return False


class UnaryBool(Bool):
    """ Lifts predicates into the DSL. """
    def __init__(self, pred):
        self.pred = pred

    def __call__(self, data):
            if not isinstance(data, list):
                data = [data]
            for d in data:
                try:
                    if self.pred(d):
                        return True
                except:
                    pass
            return False


def BinaryBool(pred):
    """ Lifts predicates that take an argument into the DSL. """
    class Predicate(Bool):
        def __init__(self, value, ignore_case=False):
            self.value = caseless(value) if ignore_case else value
            self.ignore_case = ignore_case

        def __call__(self, data):
            if not isinstance(data, list):
                data = [data]
            for d in data:
                try:
                    if pred(caseless(d) if self.ignore_case else d, self.value):
                        return True
                except:
                    pass
            return False
    return Predicate


startswith = BinaryBool(str.startswith)
istartswith = partial(startswith, ignore_case=True)

endswith = BinaryBool(str.endswith)
iendswith = partial(endswith, ignore_case=True)

contains = BinaryBool(operator.contains)
icontains = partial(contains, ignore_case=True)

le = BinaryBool(operator.le)
ile = partial(le, ignore_case=True)

lt = BinaryBool(operator.lt)
ilt = partial(lt, ignore_case=True)

ge = BinaryBool(operator.ge)
ige = partial(ge, ignore_case=True)

gt = BinaryBool(operator.gt)
igt = partial(gt, ignore_case=True)

eq = BinaryBool(operator.eq)
ieq = partial(eq, ignore_case=True)


def __make_name_pred(name):
    return name if callable(name) else lambda x: (name is None or x == name)


def __make_attrs_pred(attrs):
    if len(attrs) == 1 and callable(attrs[0]):
        return attrs[0]
    return lambda x: (len(attrs) == 0 or any(a in x for a in attrs))


def __compose(g, f):
    def inner(nodes):
        results = []
        tmp = f(nodes)
        for t in tmp:
            results.extend(g(t.children))
        return results
    return inner


# selection strategy when multiple results are returned but only one is needed
first = 0
last = -1


def select(*queries, **kwargs):
    """
    Builds a function that will execute the specified queries against a list of
    Nodes.
    """
    def make_query(*args):
        def simple_query(nodes):
            if len(args) == 0:
                return nodes
            pred = args[0]
            results = []
            if isinstance(pred, list):
                funcs = [make_query(q) for q in pred]
                return __or(funcs, nodes)
            elif isinstance(pred, tuple):
                name, attrs = pred[0], pred[1:]
                name_pred = __make_name_pred(name)
                attrs_pred = __make_attrs_pred(attrs)
                for n in nodes:
                    if name_pred(n.name) and attrs_pred(n.attrs):
                        results.append(n)
            else:
                name_pred = __make_name_pred(pred)
                for n in nodes:
                    if name_pred(n.name):
                        results.append(n)
            return results
        if len(args) > 1:
            return __compose(make_query(*args[1:]), simple_query)
        return simple_query

    def deep_query(query, nodes):
        """ Slide the query down the branches. """
        def inner(children):
            results = []
            for c in children:
                if query([c]):
                    results.append(c)
                results.extend(inner(c.children))
            return results
        return inner(nodes)

    def unique(roots):
        seen = set()
        results = []
        for r in roots:
            if r not in seen:
                seen.add(r)
                results.append(r)
        return results

    def compiled_query(nodes):
        """
        This is the compiled query that can be run against a configuration.
        """
        query = make_query(*queries)

        roots = kwargs.get("roots", True)
        if kwargs.get("deep", False):
            results = deep_query(query, nodes)
            if roots:
                results = unique([r.root for r in results])
        elif roots:
            results = unique([n.root for n in query(nodes)])
        else:
            results = query(nodes)

        one = kwargs.get("one")
        if one is None:
            return SearchResult(children=results)
        return results[one] if results else None
    return compiled_query



#!/usr/bin/env python
'''
Python bindings for libmagic
'''

import ctypes

from ctypes import *
from ctypes.util import find_library


def _init():
    """
    Loads the shared library through ctypes and returns a library
    L{ctypes.CDLL} instance
    """
    return ctypes.cdll.LoadLibrary(find_library('magic'))

_libraries = {}
_libraries['magic'] = _init()

# Flag constants for open and setflags
MAGIC_NONE = NONE = 0
MAGIC_DEBUG = DEBUG = 1
MAGIC_SYMLINK = SYMLINK = 2
MAGIC_COMPRESS = COMPRESS = 4
MAGIC_DEVICES = DEVICES = 8
MAGIC_MIME_TYPE = MIME_TYPE = 16
MAGIC_CONTINUE = CONTINUE = 32
MAGIC_CHECK = CHECK = 64
MAGIC_PRESERVE_ATIME = PRESERVE_ATIME = 128
MAGIC_RAW = RAW = 256
MAGIC_ERROR = ERROR = 512
MAGIC_MIME_ENCODING = MIME_ENCODING = 1024
MAGIC_MIME = MIME = 1040
MAGIC_APPLE = APPLE = 2048

MAGIC_NO_CHECK_COMPRESS = NO_CHECK_COMPRESS = 4096
MAGIC_NO_CHECK_TAR = NO_CHECK_TAR = 8192
MAGIC_NO_CHECK_SOFT = NO_CHECK_SOFT = 16384
MAGIC_NO_CHECK_APPTYPE = NO_CHECK_APPTYPE = 32768
MAGIC_NO_CHECK_ELF = NO_CHECK_ELF = 65536
MAGIC_NO_CHECK_TEXT = NO_CHECK_TEXT = 131072
MAGIC_NO_CHECK_CDF = NO_CHECK_CDF = 262144
MAGIC_NO_CHECK_TOKENS = NO_CHECK_TOKENS = 1048576
MAGIC_NO_CHECK_ENCODING = NO_CHECK_ENCODING = 2097152

MAGIC_NO_CHECK_BUILTIN = NO_CHECK_BUILTIN = 4173824


class magic_set(Structure):
    pass
magic_set._fields_ = []
magic_t = POINTER(magic_set)

_open = _libraries['magic'].magic_open
_open.restype = magic_t
_open.argtypes = [c_int]

_close = _libraries['magic'].magic_close
_close.restype = None
_close.argtypes = [magic_t]

_file = _libraries['magic'].magic_file
_file.restype = c_char_p
_file.argtypes = [magic_t, c_char_p]

_descriptor = _libraries['magic'].magic_descriptor
_descriptor.restype = c_char_p
_descriptor.argtypes = [magic_t, c_int]

_buffer = _libraries['magic'].magic_buffer
_buffer.restype = c_char_p
_buffer.argtypes = [magic_t, c_void_p, c_size_t]

_error = _libraries['magic'].magic_error
_error.restype = c_char_p
_error.argtypes = [magic_t]

_setflags = _libraries['magic'].magic_setflags
_setflags.restype = c_int
_setflags.argtypes = [magic_t, c_int]

_load = _libraries['magic'].magic_load
_load.restype = c_int
_load.argtypes = [magic_t, c_char_p]

_compile = _libraries['magic'].magic_compile
_compile.restype = c_int
_compile.argtypes = [magic_t, c_char_p]

_check = _libraries['magic'].magic_check
_check.restype = c_int
_check.argtypes = [magic_t, c_char_p]

_errno = _libraries['magic'].magic_errno
_errno.restype = c_int
_errno.argtypes = [magic_t]


class Magic(object):
    def __init__(self, ms):
        self._magic_t = ms

    def close(self):
        """
        Closes the magic database and deallocates any resources used.
        """
        _close(self._magic_t)

    def file(self, filename):
        """
        Returns a textual description of the contents of the argument passed
        as a filename or None if an error occurred and the MAGIC_ERROR flag
        is set.  A call to errno() will return the numeric error code.
        """
        try:  # attempt python3 approach first
            if isinstance(filename, bytes):
                bi = filename
            else:
                bi = bytes(filename, 'utf-8')
            return str(_file(self._magic_t, bi), 'utf-8')
        except:
            return _file(self._magic_t, filename.encode('utf-8'))

    def descriptor(self, fd):
        """
        Like the file method, but the argument is a file descriptor.
        """
        return _descriptor(self._magic_t, fd)

    def buffer(self, buf):
        """
        Returns a textual description of the contents of the argument passed
        as a buffer or None if an error occurred and the MAGIC_ERROR flag
        is set. A call to errno() will return the numeric error code.
        """
        try:  # attempt python3 approach first
            return str(_buffer(self._magic_t, buf, len(buf)), 'utf-8')
        except:
            return _buffer(self._magic_t, buf, len(buf))

    def error(self):
        """
        Returns a textual explanation of the last error or None
        if there was no error.
        """
        try:  # attempt python3 approach first
            return str(_error(self._magic_t), 'utf-8')
        except:
            return _error(self._magic_t)

    def setflags(self, flags):
        """
        Set flags on the magic object which determine how magic checking
        behaves; a bitwise OR of the flags described in libmagic(3), but
        without the MAGIC_ prefix.

        Returns -1 on systems that don't support utime(2) or utimes(2)
        when PRESERVE_ATIME is set.
        """
        return _setflags(self._magic_t, flags)

    def load(self, filename=None):
        """
        Must be called to load entries in the colon separated list of database
        files passed as argument or the default database file if no argument
        before any magic queries can be performed.

        Returns 0 on success and -1 on failure.
        """
        return _load(self._magic_t, filename)

    def compile(self, dbs):
        """
        Compile entries in the colon separated list of database files
        passed as argument or the default database file if no argument.
        Returns 0 on success and -1 on failure.
        The compiled files created are named from the basename(1) of each file
        argument with ".mgc" appended to it.
        """
        return _compile(self._magic_t, dbs)

    def check(self, dbs):
        """
        Check the validity of entries in the colon separated list of
        database files passed as argument or the default database file
        if no argument.
        Returns 0 on success and -1 on failure.
        """
        return _check(self._magic_t, dbs)

    def errno(self):
        """
        Returns a numeric error code. If return value is 0, an internal
        magic error occurred. If return value is non-zero, the value is
        an OS error code. Use the errno module or os.strerror() can be used
        to provide detailed error information.
        """
        return _errno(self._magic_t)


def open(flags):
    """
    Returns a magic object on success and None on failure.
    Flags argument as for setflags.
    """
    return Magic(_open(flags))
